import os

from px import px_meminfo


def test_get_ram_numbers_live():
    info = px_meminfo._get_ram_numbers()
    assert info is not None

    total_bytes, wanted_bytes = info
    assert total_bytes > 0
    assert wanted_bytes > 0


def test_get_ram_numbers_in_swedish():
    os.environ["LANG"] = "sv_SE.UTF-8"
    os.environ["LC_TIME"] = "sv_SE.UTF-8"
    os.environ["LC_NUMERIC"] = "sv_SE.UTF-8"

    test_get_ram_numbers_live()


def test_get_ram_numbers_from_proc_none():
    my_dir = os.path.dirname(__file__)

    info = px_meminfo._get_ram_numbers_from_proc(
        os.path.join(my_dir, "file name that doesn't exist")
    )
    assert info is None


def test_get_ram_numbers_from_proc_2010():
    my_dir = os.path.dirname(__file__)

    info = px_meminfo._get_ram_numbers_from_proc(
        os.path.join(my_dir, "proc-meminfo-2010.txt")
    )
    assert info is not None

    total_bytes, wanted_bytes = info
    assert total_bytes == 1921988 * 1024

    # From before MemAvailable's invention:
    # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=34e431b0ae398fc54ea69ff85ec700722c9da773
    available_kb = 1374408 + 32688 + 370540 + 0
    assert wanted_bytes == total_bytes - available_kb * 1024


def test_get_ram_numbers_from_proc_2020():
    my_dir = os.path.dirname(__file__)

    info = px_meminfo._get_ram_numbers_from_proc(
        os.path.join(my_dir, "proc-meminfo-2020.txt")
    )
    assert info is not None

    total_bytes, wanted_bytes = info
    assert total_bytes == 371474308 * 1024

    # Use MemAvailable when available:
    # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?id=34e431b0ae398fc54ea69ff85ec700722c9da773
    available_kb = 294625460
    assert wanted_bytes == total_bytes - available_kb * 1024



import os

from px import px_cpuinfo


def test_get_core_count():
    physical, logical = px_cpuinfo.get_core_count()

    assert physical >= 1
    assert logical >= physical
    assert logical % physical == 0


def assert_core_counts_from_file(filename, expected_physical, expected_logical):
    my_dir = os.path.dirname(__file__)

    info = px_cpuinfo.get_core_count_from_proc_cpuinfo(os.path.join(my_dir, filename))
    assert info
    physical, logical = info
    assert physical == expected_physical
    assert logical == expected_logical


def test_get_core_count_from_proc_cpuinfo():
    assert_core_counts_from_file("proc-cpuinfo-1p1l.txt", 1, 1)
    assert_core_counts_from_file("proc-cpuinfo-2p4l.txt", 2, 4)
    assert_core_counts_from_file("proc-cpuinfo-24p96l.txt", 24, 96)

    # This one is from my cell phone, just to provide a weird corner case example of things
    # we may have to handle.
    assert_core_counts_from_file("proc-cpuinfo-8p8l.txt", 8, 8)

    assert px_cpuinfo.get_core_count_from_proc_cpuinfo("/does/not/exist") is None


def test_get_core_count_from_sysctl():
    test_me = px_cpuinfo.get_core_count_from_sysctl()
    if test_me is None:
        # We're likely on Linux
        return

    # Sanity check, since we don't know the answer this is the best we can do
    physical, logical = test_me
    assert physical >= 1
    assert logical >= physical
    assert logical % physical == 0


def test_parse_sysctl_output():
    result = px_cpuinfo.parse_sysctl_output([])
    assert result == (None, None)

    result = px_cpuinfo.parse_sysctl_output(["gurka"])
    assert result == (None, None)

    result = px_cpuinfo.parse_sysctl_output(["hw.physicalcpu: 1", "hw.logicalcpu: 2"])
    assert result == (1, 2)

    result = px_cpuinfo.parse_sysctl_output(["hw.physicalcpu: 12", "hw.logicalcpu: 34"])
    assert result == (12, 34)



from px import px_load
from px import px_terminal


def test_averages_to_levels():
    l0, l1, l2, peak = px_load.averages_to_levels(0.0, 0.0, 0.0)
    assert (l0, l1, l2, peak) == (0, 0, 0, 1.0)

    # Levels are:
    # 0: 0/6 - 1/6
    # 1: 1/6 - 3/6
    # 2: 3/6 - 5/6
    # 3: 5/6 - 6/6
    l0, l1, l2, peak = px_load.averages_to_levels(
        1.0 / 6.0 - 0.01, 1.0 / 6.0 + 0.01, 1.0
    )
    assert (l0, l1, l2, peak) == (0, 1, 3, 1.0)

    l0, l1, l2, peak = px_load.averages_to_levels(
        5.0 / 6.0 - 0.01, 5.0 / 6.0 + 0.01, 1.0
    )
    assert (l0, l1, l2, peak) == (2, 3, 3, 1.0)


def test_levels_to_graph():
    # Levels 0-3 should be visualized with 1-4 dots. And each character should
    # contain two bars. So this test case wants: "1, 2", "3, 4".
    assert px_load.levels_to_graph([0, 1, 2, 3]) == "⣠⣾"

    # Uneven-numbered level arrays should be padded with an empty column on the
    # left
    assert px_load.levels_to_graph([0, 1, 3]) == "⢀⣼"

    assert px_load.levels_to_graph([]) == ""
    assert len(px_load.levels_to_graph([0])) == 1
    assert len(px_load.levels_to_graph([1, 2])) == 1
    assert len(px_load.levels_to_graph([3, 0, 1])) == 2
    assert len(px_load.levels_to_graph([1] * 15)) == 8


def test_get_load_string():
    px_terminal._enable_color = True
    CSI = "\x1b["
    assert "0.3" + CSI in px_load.get_load_string((0.3, 0.2, 0.1))
    assert "3.0" + CSI in px_load.get_load_string((3.0, 0.2, 0.1))
    assert "1.1" + CSI in px_load.get_load_string((1.135135, 0.2, 0.1))
    assert "2.0" + CSI in px_load.get_load_string((2.0, 3.0, 4.0))



import os.path

from px import px_file
from px import px_ipc_map
from . import testutils


def get_other_end_pids(my_file, all_files):
    """Wrapper around IpcMap._get_other_end_pids() so that we can test it"""
    return testutils.create_ipc_map(my_file.pid, all_files)._get_other_end_pids(my_file)


def test_get_other_end_pids_basic():
    PIPE_ID = "0x919E1D"
    files = [testutils.create_file("PIPE", "name", PIPE_ID, 25)]

    my_end = testutils.create_file("PIPE", "[] " + PIPE_ID, "0x1234", 42)
    found = get_other_end_pids(my_end, files)
    assert found == {25}

    my_end = testutils.create_file("PIPE", "[] ->" + PIPE_ID, "0x1234", 42)
    found = get_other_end_pids(my_end, files)
    assert found == {25}

    my_end = testutils.create_file("PIPE", "doesn't exist", "0x1234", 42)
    found = get_other_end_pids(my_end, files)
    assert found == set()


def test_get_other_end_pids_osx_pipe1():
    files = [
        testutils.create_file("PIPE", "[] ->0xAda", "0xE0e", 25),
        testutils.create_file("PIPE", "[] ->0x59aee", "0xE0e", 26),
    ]
    my_end = testutils.create_file("PIPE", "[] ->0xE0e", "0x8000", 42)
    found = get_other_end_pids(my_end, files)
    assert found == {25, 26}


def test_get_other_end_pids_osx_pipe2():
    files = [
        testutils.create_file("PIPE", "[] ->0xAda", "0xE0e", 25),
        testutils.create_file("PIPE", "[] ->0xAda", "0x38ee", 26),
    ]
    my_end = testutils.create_file("PIPE", "[] ->0x8a8de9", "0xAda", 42)
    found = get_other_end_pids(my_end, files)
    assert found == {25, 26}


def test_get_other_end_pids_linux_pipe():
    files = [
        testutils.create_file("FIFO", "pipe", None, 100, inode="100200", access="r"),
        testutils.create_file("FIFO", "pipe", None, 200, inode="100200", access="w"),
        testutils.create_file("FIFO", "pipe", None, 300, inode="100300", access="w"),
    ]
    assert get_other_end_pids(files[0], files) == {200}
    assert get_other_end_pids(files[1], files) == {100}
    assert get_other_end_pids(files[2], files) == set()


def test_get_other_end_pids_fifo1():
    files = [testutils.create_file("FIFO", "[] /fifo/name", None, 25, "r")]

    my_end = testutils.create_file("FIFO", "[] /fifo/name", None, 42, "w")
    found = get_other_end_pids(my_end, files)
    assert found == {25}


def test_get_other_end_pids_fifo2():
    files = [testutils.create_file("FIFO", "[] /fifo/name", None, 25, "w")]

    my_end = testutils.create_file("FIFO", "[] /fifo/name", None, 42, "r")
    found = get_other_end_pids(my_end, files)
    assert found == {25}


def test_get_other_end_pids_linux_socket():
    files = [
        testutils.create_file("unix", "[] socket", "0xabc123", 25, "u"),
        testutils.create_file("unix", "[] socket", "0xabc123", 26, "u"),
        testutils.create_file("unix", "[] socket", "0xdef456", 27, "u"),
    ]

    my_end = testutils.create_file("unix", "[] socket", "0xabc123", 42, "u")
    found = get_other_end_pids(my_end, files)
    assert found == {25, 26}

    # Should get all PIDs including our own
    my_end.pid = 25
    found = get_other_end_pids(my_end, files)
    assert found == {25, 26}


def test_get_other_end_pids_osx_socket():
    """This is from a real world example"""
    atom_file = testutils.create_file(
        "unix", "->0xebb7d964ac3da0b7", "0xebb7d964c20e6947", 1234, "u"
    )
    python_file = testutils.create_file(
        "unix", "->0xebb7d964c20e6947", "0xebb7d964ac3da0b7", 4567, "u"
    )
    files = [atom_file, python_file]

    assert 4567 in get_other_end_pids(atom_file, files)
    assert 1234 in get_other_end_pids(python_file, files)


def test_get_other_end_pids_localhost_socket():
    # Real world test data
    tc_file = testutils.create_file(
        "IPv4", "localhost:33815->localhost:postgresql", "444139298", 33019, "u"
    )
    postgres_file = testutils.create_file(
        "IPv4", "localhost:postgresql->localhost:33815", "444206166", 42745, "u"
    )
    files = [tc_file, postgres_file]

    tc_ipc_map = testutils.create_ipc_map(33019, files)
    assert 42745 in tc_ipc_map._get_other_end_pids(tc_file)
    assert tc_file not in tc_ipc_map.network_connections

    postgres_ipc_map = testutils.create_ipc_map(42745, files)
    assert 33019 in postgres_ipc_map._get_other_end_pids(postgres_file)
    assert postgres_file not in postgres_ipc_map.network_connections


def test_get_other_end_pids_localhost_socket_names():
    # lsof usually presents a number of different names for localhost, because
    # of different network interfaces and other reasons. Make sure we identify
    # those and treat them like localhost.
    tc_file = testutils.create_file(
        "IPv4", "127.0.0.42:33815->localhost:postgresql", "444139298", 33019, "u"
    )
    postgres_file = testutils.create_file(
        "IPv4", "localhost:postgresql->127.0.0.42:33815", "444206166", 42745, "u"
    )
    files = [tc_file, postgres_file]

    tc_ipc_map = testutils.create_ipc_map(33019, files)
    assert 42745 in tc_ipc_map._get_other_end_pids(tc_file)
    assert tc_file not in tc_ipc_map.network_connections

    postgres_ipc_map = testutils.create_ipc_map(42745, files)
    assert 33019 in postgres_ipc_map._get_other_end_pids(postgres_file)
    assert postgres_file not in postgres_ipc_map.network_connections


def test_get_ipc_map_1():
    """Tyre kick IpcMap with some real world data"""
    files = None
    my_dir = os.path.dirname(__file__)
    with open(
        os.path.join(my_dir, "lsof-test-output-linux-1.txt"), encoding="utf-8"
    ) as lsof_output:
        files = px_file.lsof_to_files(lsof_output.read())

    ipc_map = testutils.create_ipc_map(1997, files)
    keys = list(ipc_map.keys())
    assert len(keys) == 2

    peer0 = keys[0]
    assert len(ipc_map[peer0]) == 1

    peer1 = keys[1]
    assert len(ipc_map[peer1]) == 1


def test_get_ipc_map_2():
    """Tyre kick IpcMap with some real world data"""
    files = None
    my_dir = os.path.dirname(__file__)
    with open(
        os.path.join(my_dir, "lsof-test-output-linux-2.txt"), encoding="utf-8"
    ) as lsof_output:
        files = px_file.lsof_to_files(lsof_output.read())

    ipc_map = testutils.create_ipc_map(777, files)
    keys = list(ipc_map.keys())
    assert len(keys) == 1

    peer0 = keys[0]
    assert len(ipc_map[peer0]) == 4


def test_stdfds_base():
    files = [
        testutils.create_file("REG", "/wherever/stdin", None, 1234, fd=0),
        testutils.create_file("REG", "/wherever/stdout", None, 1234, fd=1),
        testutils.create_file("REG", "/wherever/stderr", None, 1234, fd=2),
    ]

    ipc_map = testutils.create_ipc_map(1234, files)
    assert ipc_map.fds[0] == "/wherever/stdin"
    assert ipc_map.fds[1] == "/wherever/stdout"
    assert ipc_map.fds[2] == "/wherever/stderr"


def test_stdfds_closed():
    files = [
        testutils.create_file("REG", "/wherever", None, 1234, fd=3),
    ]

    ipc_map = testutils.create_ipc_map(1234, files)

    # If we get other fds but not 0-2, we assume 0-2 to have been closed.
    assert ipc_map.fds[0] == "<closed>"
    assert ipc_map.fds[1] == "<closed>"
    assert ipc_map.fds[2] == "<closed>"


def test_stdfds_unavailable():
    ipc_map = testutils.create_ipc_map(1234, [])

    # If we get no fds at all for this process, we assume lack of data
    assert ipc_map.fds[0] == "<unavailable, running px as root might help>"
    assert ipc_map.fds[1] == "<unavailable, running px as root might help>"
    assert ipc_map.fds[2] == "<unavailable, running px as root might help>"


def test_stdfds_ipc_and_network():
    files = [
        # Set up stdin to do a network connection to another local process
        testutils.create_file(
            "IPv4",
            "localhost:33815->localhost:postgresql",
            "444139298",
            1234,
            "u",
            fd=0,
        ),
        testutils.create_file(
            "IPv4", "localhost:postgresql->localhost:33815", "444206166", 1000, "u"
        ),
        # Set up stdout to do a network connection to 8.8.8.8
        testutils.create_file(
            "IPv4", "127.0.0.1:9999->8.8.8.8:https", None, 1234, fd=1
        ),
        # Set up stderr to pipe to another process. Pipes are real-world OS X ones.
        testutils.create_file(
            "PIPE", "->0x3922f6866312c495", "0x3922f6866312cb55", 1234, fd=2
        ),
        testutils.create_file(
            "PIPE", "->0x3922f6866312cb55", "0x3922f6866312c495", 1002
        ),
    ]

    ipc_map = testutils.create_ipc_map(1234, files)
    assert (
        ipc_map.fds[0]
        == "[IPv4] -> cupsd(1000) (localhost:33815->localhost:postgresql)"
    )
    assert ipc_map.fds[1] in [
        "[IPv4] localhost:9999->google-public-dns-a.google.com:https",
        "[IPv4] localhost:9999->dns.google:https",
    ]
    assert ipc_map.fds[2] == "[PIPE] -> cupsd(1002) (0x3922f6866312c495)"


def test_stdfds_pipe_to_unknown_not_root():
    files = [
        # Set up stderr as an unconnected pipe. The pipes is a real-world OS X one.
        testutils.create_file(
            "PIPE", "->0x3922f6866312c495", "0x3922f6866312cb55", 1234, fd=2
        ),
    ]

    ipc_map = testutils.create_ipc_map(1234, files, is_root=False)
    assert (
        ipc_map.fds[2]
        == "[PIPE] <destination not found, try running px as root> (0x3922f6866312c495)"
    )


def test_stdfds_pipe_to_unknown_is_root():
    files = [
        # Set up stderr as an unconnected pipe. The pipe is a real-world OS X one.
        testutils.create_file(
            "PIPE", "->0x3922f6866312c495", "0x3922f6866312cb55", 1234, fd=2
        ),
    ]

    ipc_map = testutils.create_ipc_map(1234, files, is_root=True)
    assert ipc_map.fds[2] == "[PIPE] <not connected> (0x3922f6866312c495)"


def test_stdfds_osx_pipe_to_unknown_is_root():
    files = [
        # Set up stdin as an unconnected pipe. The pipe is a real-world OS X one.
        testutils.create_file("PIPE", "", "0x3922f6866312cb55", 1234, fd=0),
    ]

    ipc_map = testutils.create_ipc_map(1234, files, is_root=True)
    assert ipc_map.fds[0] == "[PIPE] <not connected> (0x3922f6866312cb55)"


def test_ipc_pipe_osx():
    # These pipes are real-world OS X ones
    f1 = testutils.create_file(
        "PIPE", "->0x3922f6866312c495", "0x3922f6866312cb55", 2222
    )
    assert f1.fifo_id() == "->0x3922f6866312c495"

    f2 = testutils.create_file(
        "PIPE", "->0x3922f6866312cb55", "0x3922f6866312c495", 1001
    )
    assert f2.fifo_id() == "->0x3922f6866312cb55"

    files = [f1, f2]

    ipc_map = testutils.create_ipc_map(2222, files)
    assert 1001 in ipc_map._get_other_end_pids(f1)
    assert 2222 in ipc_map._get_other_end_pids(f2)
    assert len(list(ipc_map.keys())) == 1


def test_peer_process_str():
    assert str(px_ipc_map.PeerProcess(pid=45)) == "PID 45"
    assert str(px_ipc_map.PeerProcess(name="Johan")) == "Johan"



from . import testutils

from px import px_ioload


def test_parse_netstat_ib_output():
    sample_netstat_ib_output = testutils.load("netstat-ib.txt")

    expected = [
        px_ioload.Sample("lo0 incoming", 903920),
        px_ioload.Sample("lo0 outgoing", 903920),
        px_ioload.Sample("en0 incoming", 1410907682),
        px_ioload.Sample("en0 outgoing", 67569820),
        px_ioload.Sample("awdl0 incoming", 0),
        px_ioload.Sample("awdl0 outgoing", 11760),
        px_ioload.Sample("utun0 incoming", 0),
        px_ioload.Sample("utun0 outgoing", 200),
        px_ioload.Sample("utun1 incoming", 0),
        px_ioload.Sample("utun1 outgoing", 200),
    ]
    actual = px_ioload.parse_netstat_ib_output(sample_netstat_ib_output)

    assert actual == expected


def test_parse_iostat_output():
    sample_iostat_output = testutils.load("iostat-dki-n-99.txt")

    expected = [px_ioload.Sample("disk0", 46066166661)]
    actual = px_ioload.parse_iostat_output(sample_iostat_output)

    assert actual == expected


def test_parse_proc_net_dev():
    proc_net_dev_contents = testutils.load("proc-net-dev.txt")

    # No data expected for the all-zero interfaces lo, tunl0 and ip6tnl0
    expected = [
        px_ioload.Sample("eth0 incoming", 29819439),
        px_ioload.Sample("eth0 outgoing", 364327),
    ]
    actual = px_ioload.parse_proc_net_dev(proc_net_dev_contents)

    assert actual == expected


def test_parse_proc_diskstats():
    proc_diskstats_contents = testutils.load("proc-diskstats.txt")

    # No data expected for the all-zero disks ram*, loop* and nbd*. Get stats
    # per partition (name ends in a number), not per drive.
    expected = [
        px_ioload.Sample("vda1 read", 804910 * 512),
        px_ioload.Sample("vda1 write", 1852480 * 512),
    ]
    actual = px_ioload.parse_proc_diskstats(proc_diskstats_contents)

    assert actual == expected


def test_system_state():
    system_state = px_ioload.SystemState()

    # We should get *something*
    assert system_state.samples



from px import px_tree
from px import px_terminal
from px import px_process

from . import testutils

from typing import List
import datetime


def resolve(processes: List[px_process.PxProcess]) -> List[px_process.PxProcess]:
    now = datetime.datetime.now(datetime.timezone.utc).astimezone()

    process_dict = {p.pid: p for p in processes}
    px_process.resolve_links(process_dict, now)

    return processes


def test_empty():
    assert px_tree._generate_tree([], "") == []


def test_parent_child():
    assert px_tree._generate_tree(
        resolve(
            [
                testutils.create_process(pid=1, ppid=0, commandline="parent"),
                testutils.create_process(pid=2, ppid=1, commandline="child"),
            ]
        ),
        "",
    ) == ["parent(1)", "  child(2)"]


def test_search():
    assert px_tree._generate_tree(
        resolve(
            [
                testutils.create_process(pid=1, ppid=0, commandline="root"),
                testutils.create_process(pid=2, ppid=1, commandline="find-me"),
                testutils.create_process(pid=3, ppid=1, commandline="uninteresting"),
            ]
        ),
        "find-me",
    ) == ["root(1)", f"  {px_terminal.bold('find-me')}(2)"]


def test_no_search():
    assert px_tree._generate_tree(
        resolve(
            [
                testutils.create_process(pid=1, ppid=0, commandline="root"),
                testutils.create_process(pid=2, ppid=1, commandline="find-me"),
                testutils.create_process(pid=3, ppid=1, commandline="uninteresting"),
            ]
        ),
        "",
    ) == ["root(1)", "  find-me(2)", "  uninteresting(3)"]


def test_search_no_matches():
    assert (
        px_tree._generate_tree(
            resolve(
                [
                    testutils.create_process(pid=1, ppid=0, commandline="root"),
                    testutils.create_process(pid=2, ppid=1, commandline="find-me"),
                    testutils.create_process(
                        pid=3, ppid=1, commandline="uninteresting"
                    ),
                ]
            ),
            "not-there",
        )
        == []
    )


def test_coalesce_no_search():
    assert px_tree._generate_tree(
        resolve(
            [
                testutils.create_process(pid=1, ppid=0, commandline="root"),
                testutils.create_process(pid=2, ppid=1, commandline="process"),
                testutils.create_process(pid=3, ppid=1, commandline="process"),
            ]
        ),
        "",
    ) == ["root(1)", f"  process... ({px_terminal.bold('2×')})"]


def test_coalesce_with_search():
    assert px_tree._generate_tree(
        resolve(
            [
                testutils.create_process(pid=1, ppid=0, commandline="root"),
                testutils.create_process(pid=2, ppid=1, commandline="process"),
                testutils.create_process(pid=3, ppid=1, commandline="process"),
            ]
        ),
        "process",
    ) == [
        "root(1)",
        f"  {px_terminal.bold('process')}(2)",
        f"  {px_terminal.bold('process')}(3)",
    ]


def test_coalesce_no_leaf():
    assert px_tree._generate_tree(
        resolve(
            [
                testutils.create_process(pid=1, ppid=0, commandline="root"),
                testutils.create_process(pid=2, ppid=1, commandline="process"),
                testutils.create_process(pid=3, ppid=1, commandline="process"),
                testutils.create_process(pid=4, ppid=1, commandline="process"),
                testutils.create_process(pid=5, ppid=1, commandline="process"),
                testutils.create_process(pid=6, ppid=1, commandline="process"),
                testutils.create_process(pid=7, ppid=4, commandline="subprocess"),
            ]
        ),
        "",
    ) == [
        "root(1)",
        f"  process... ({px_terminal.bold('4×')})",
        "  process(4)",
        "    subprocess(7)",
    ]


def test_coalescer():
    test_me = px_tree.Coalescer(0)
    assert not test_me.submit(testutils.create_process(pid=1, commandline="Kuala"), "")
    assert test_me.submit(testutils.create_process(commandline="Lumpur"), "") == [
        "Kuala(1)"
    ]
    assert not test_me.submit(testutils.create_process(commandline="Lumpur"), "")
    assert not test_me.submit(testutils.create_process(commandline="Lumpur"), "")
    assert test_me.flush() == [f"Lumpur... ({px_terminal.bold('3×')})"]



from px import px_terminal
from px import px_launchcounter

from . import testutils


def test_list_new_launches():
    process = testutils.create_process(pid=100, timestring="Mon Apr  7 09:33:11 2010")
    process_identical = testutils.create_process(
        pid=100, timestring="Mon Apr  7 09:33:11 2010"
    )
    process_other_pid = testutils.create_process(
        pid=101, timestring="Mon Apr  7 09:33:11 2010"
    )
    process_other_starttime = testutils.create_process(
        pid=100, timestring="Mon Apr  8 09:33:11 2010"
    )

    before = [process]
    after = [process_identical, process_other_pid, process_other_starttime]

    new_processes = px_launchcounter._list_new_launches(before, after)

    # We should get unicode responses from getch()
    assert new_processes == [process_other_pid, process_other_starttime]


def test_get_screen_lines_coalesces():
    px_terminal._enable_color = True
    # If we have both "init"->"iTerm" and "init"->"iTerm"->"fish",
    # they should be reported as just "init"->"iTerm"->"fish".
    launchcounter = px_launchcounter.Launchcounter()
    launchcounter._register_launches(
        [
            testutils.fake_callchain("init", "iTerm"),
            testutils.fake_callchain("init", "iTerm", "fish"),
        ]
    )
    lines = launchcounter.get_screen_lines()
    assert lines == [
        "init -> "
        + px_terminal.bold("iTerm")
        + "(1) -> "
        + px_terminal.bold("fish")
        + "(1)"
    ]

    # Then the same thing backwards
    launchcounter = px_launchcounter.Launchcounter()
    launchcounter._register_launches(
        [
            testutils.fake_callchain("init", "iTerm", "fish"),
            testutils.fake_callchain("init", "iTerm"),
        ]
    )
    lines = launchcounter.get_screen_lines()
    assert lines == [
        "init -> "
        + px_terminal.bold("iTerm")
        + "(1) -> "
        + px_terminal.bold("fish")
        + "(1)"
    ]


def test_print_launch_counts():
    px_terminal._enable_color = True
    launchcounter = px_launchcounter.Launchcounter()
    launchcounter._register_launches(
        [
            testutils.fake_callchain("init", "iTerm"),
            testutils.fake_callchain("init", "iTerm", "fish"),
            testutils.fake_callchain("init", "iTerm", "fish"),
            testutils.fake_callchain("init", "iTerm"),
            testutils.fake_callchain("init", "iTerm"),
        ]
    )
    lines = launchcounter.get_screen_lines()
    assert lines == [
        "init -> "
        + px_terminal.bold("iTerm")
        + "(3) -> "
        + px_terminal.bold("fish")
        + "(2)"
    ]


def test_ignore_surrounding_parentheses():
    px_terminal._enable_color = True
    launchcounter = px_launchcounter.Launchcounter()
    launchcounter._register_launches(
        [
            testutils.fake_callchain("init", "iTerm"),
            testutils.fake_callchain("init", "iTerm"),
            testutils.fake_callchain("init", "(iTerm)"),
            testutils.fake_callchain("init", "iTerm()"),
            testutils.fake_callchain("init", "i(Term)"),
            testutils.fake_callchain("init", "(i)Term"),
        ]
    )
    lines = launchcounter.get_screen_lines()

    # Note that the ordering of the lines doesn't really matter here, just as
    # long as all of them are in there
    assert set(lines) == {
        "init -> " + px_terminal.bold("iTerm") + "(3)",
        "init -> " + px_terminal.bold("(i)Term") + "(1)",
        "init -> " + px_terminal.bold("i(Term)") + "(1)",
        "init -> " + px_terminal.bold("iTerm()") + "(1)",
    }


def test_to_tuple_list():
    assert px_launchcounter._to_tuple_list(("a", "b", "c"), 5) == [
        ("a", 0),
        ("b", 0),
        ("c", 5),
    ]

    assert px_launchcounter._to_tuple_list(("a",), 5) == [("a", 5)]


def test_merge_tuple_lists():
    l1 = [("a", 0), ("b", 2), ("c", 5)]
    l2 = [("a", 0), ("b", 4)]
    l3 = [("a", 0), ("e", 4)]

    assert px_launchcounter._merge_tuple_lists(l1, l2) == [("a", 0), ("b", 6), ("c", 5)]

    assert px_launchcounter._merge_tuple_lists(l1, l3) is None


def test_sort_launchers_lists():
    l1 = [("a", 3), ("b", 4)]
    l2 = [("a", 5), ("e", 0)]

    # List with highest individual number should come first
    assert px_launchcounter.sort_launchers_list([l1, l2]) == [l2, l1]
    assert px_launchcounter.sort_launchers_list([l2, l1]) == [l2, l1]



import os

from px import px_top
from px import px_poller
from px import px_process
from px import px_terminal
from px import px_launchcounter

from . import testutils


def test_adjust_cpu_times():
    now = testutils.local_now()

    current = [
        px_process.create_kernel_process(now),
        testutils.create_process(
            pid=100, cputime="0:10.00", commandline="only in current"
        ),
        testutils.create_process(
            pid=200,
            cputime="0:20.00",
            commandline="re-used PID baseline",
            timestring="Mon May  7 09:33:11 2010",
        ),
        testutils.create_process(
            pid=300, cputime="0:30.00", commandline="relevant baseline"
        ),
    ]
    baseline = {
        0: (current[0].start_time, 0),
        200: (current[2].start_time, 2.0),
        300: (current[3].start_time, 3.0),
        400: (now, 3.0),
    }

    actual = px_process.order_best_last(px_top.adjust_cpu_times(baseline, current))
    expected = px_process.order_best_last(
        [
            px_process.create_kernel_process(now),
            testutils.create_process(
                pid=100, cputime="0:10.00", commandline="only in current"
            ),
            testutils.create_process(
                pid=200,
                cputime="0:18.00",
                commandline="re-used PID baseline",
                timestring="Mon May  7 09:33:11 2010",
            ),
            testutils.create_process(
                pid=300, cputime="0:27.00", commandline="relevant baseline"
            ),
        ]
    )

    assert actual == expected


def test_get_toplist():
    current = px_process.get_all()
    baseline = {p.pid: (p.start_time, p.cpu_time_seconds or 0.0) for p in current}
    toplist = px_top.get_toplist(baseline, px_process.get_all())
    for process in toplist:
        assert process.aggregated_cpu_time_seconds is not None
        assert process.aggregated_cpu_time_s != "--"


def test_get_command():
    pipe = os.pipe()
    read, write = pipe
    os.write(write, b"q")

    assert px_top.get_command(timeout_seconds=0, fd=read) == px_top.CMD_QUIT


def test_sigwinch_handler():
    # Args ignored at the time of writing this, fill in better values if needed
    px_terminal.sigwinch_handler(None, None)

    # sys.stdin doesn't work when STDIN has been redirected (as during testing),
    # so we need to explicitly use the STDIN fd here. Try removing it and you'll
    # see :).
    STDIN = 0
    assert px_top.get_command(timeout_seconds=0, fd=STDIN) == px_top.CMD_RESIZE


def test_redraw():
    # Just make sure it doesn't crash
    baseline = px_process.get_all()
    poller = px_poller.PxPoller()
    px_top.redraw(baseline, poller, 100, 10)


def test_get_screen_lines_low_screen():
    baseline = px_process.get_all()
    poller = px_poller.PxPoller()

    # We have to make up some number for "How low screens can we cope with?".
    # Here's the number I made up.
    SCREEN_ROWS = 11

    px_terminal._enable_color = True
    lines = px_top.get_screen_lines(baseline, poller, SCREEN_ROWS, 99)

    # Top row should contain ANSI escape codes
    CSI = "\x1b["
    assert "CSI" in lines[0].replace(CSI, "CSI")

    assert len(lines) == SCREEN_ROWS

    # Last line should be decorated
    assert "CSI" in lines[-1].replace(CSI, "CSI")


def test_get_screen_lines_high_screen():
    baseline = px_process.get_all()
    poller = px_poller.PxPoller()

    SCREEN_ROWS = 100
    px_terminal._enable_color = True
    lines = px_top.get_screen_lines(baseline, poller, SCREEN_ROWS, 99)

    # Top row should contain ANSI escape codes
    CSI = "\x1b["
    assert "CSI" in lines[0].replace(CSI, "CSI")

    assert len(lines) == SCREEN_ROWS

    # Last line should be decorated
    assert "CSI" in lines[-1].replace(CSI, "CSI")


def test_get_screen_lines_with_many_launches():
    baseline = px_process.get_all()
    launchcounter = px_launchcounter.Launchcounter()

    for i in range(1, 100):
        launchcounter._register_launches(
            [testutils.fake_callchain("init", "a" + str(i))]
        )

    poller = px_poller.PxPoller()
    poller._launchcounter = launchcounter
    poller._launchcounter_screen_lines = launchcounter.get_screen_lines()

    SCREEN_ROWS = 100
    lines = px_top.get_screen_lines(baseline, poller, SCREEN_ROWS, 99)

    assert len(lines) == SCREEN_ROWS


def test_get_screen_lines_returns_enough_lines():
    baseline = px_process.get_all()
    poller = px_poller.PxPoller()

    SCREEN_ROWS = 100000
    lines = px_top.get_screen_lines(baseline, poller, SCREEN_ROWS, 99)

    assert len(lines) == SCREEN_ROWS



import re

from px import px_file

from typing import List


def test_lsof_to_files():
    lsof = ""

    lsof += "\0".join(["p123", "\n"])
    lsof += "\0".join(["fcwd", "a ", "tDIR", "n/", "\n"])
    lsof += "\0".join(["f5", "ar", "tREG", "ncontains\nnewline", "\n"])
    lsof += "\0".join(["f6", "aw", "tREG", "d0x42", "n/somefile", "\n"])
    lsof += "\0".join(["p456", "\n"])
    lsof += "\0".join(["f7", "au", "tREG", "n/someotherfile", "\n"])
    lsof += "\0".join(["f7", "a ", "n(revoked)", "\n"])

    files = px_file.lsof_to_files(lsof)

    assert len(files) == 5

    assert files[0].pid == 123
    assert files[0].access is None
    assert files[0].device is None
    assert files[0].device_number() is None
    assert files[0].type == "DIR"
    assert files[0].name == "/"
    assert str(files[0]) == "[DIR] /"

    assert files[1].pid == 123
    assert files[1].access == "r"
    assert files[1].device is None
    assert files[1].device_number() is None
    assert files[1].type == "REG"
    assert files[1].name == "contains\nnewline"
    assert str(files[1]) == "contains\nnewline"

    assert files[2].pid == 123
    assert files[2].access == "w"
    assert files[2].device == "0x42"
    assert files[2].device_number() == 0x42
    assert files[2].type == "REG"
    assert files[2].name == "/somefile"
    assert str(files[2]) == "/somefile"

    assert files[3].pid == 456
    assert files[3].access == "rw"
    assert files[3].device is None
    assert files[3].device_number() is None
    assert files[3].type == "REG"
    assert files[3].name == "/someotherfile"
    assert str(files[3]) == "/someotherfile"

    assert files[4].pid == 456
    assert files[4].access is None
    assert files[4].device is None
    assert files[4].device_number() is None
    assert files[4].type == "??"
    assert files[4].name == "(revoked)"
    assert str(files[4]) == "[??] (revoked)"


def test_get_all():
    files = px_file.get_all()

    # As non-root I get 6000 on my system, 100 should be fine anywhere. And if
    # not, we'll just have to document our finding and lower this value
    assert len(files) > 100

    cwd_count = 0
    for file in files:
        if file.fdtype == "cwd":
            cwd_count += 1
    assert cwd_count > 0


def lsof_to_file(shard_array: List[str]) -> px_file.PxFile:
    return px_file.lsof_to_files("\0".join(shard_array + ["\n"]))[0]


def test_listen_name():
    file = lsof_to_file(["p123", "f6", "au", "tIPv4", "d0x42", "nlocalhost:63342"])
    assert file.name == "localhost:63342"
    assert str(file) == "[IPv4] localhost:63342 (LISTEN)"

    file = lsof_to_file(["p123", "f6", "au", "tIPv6", "d0x42", "nlocalhost:63342"])
    assert file.name == "localhost:63342"
    assert str(file) == "[IPv6] localhost:63342 (LISTEN)"


def test_setability():
    # Can files be stored in sets?
    a = lsof_to_file(["p123", "f6", "aw", "tREG", "d0x42", "n/somefile"])
    b = lsof_to_file(["p123", "f6", "aw", "tREG", "d0x42", "n/somefile"])
    s = {a, b}
    assert len(s) == 1


def test_local_endpoint():
    local_endpoint = lsof_to_file(
        ["p123", "f6", "au", "tIPv4", "d0x42", "nlocalhost:postgres->localhost:33331"]
    ).get_endpoints()[0]
    assert local_endpoint == "localhost:postgres"

    local_endpoint = lsof_to_file(
        ["p123", "f6", "au", "tIPv6", "d0x42", "nlocalhost:39252->localhost:39252"]
    ).get_endpoints()[0]
    assert local_endpoint == "localhost:39252"

    assert (
        lsof_to_file(
            ["p123", "f6", "au", "tIPv6", "d0x42", "nlocalhost:19091"]
        ).get_endpoints()[0]
        == "localhost:19091"
    )
    assert (
        lsof_to_file(
            ["p123", "f6", "au", "tIPv4", "d0x42", "nlocalhost:ipp (LISTEN)"]
        ).get_endpoints()[0]
        == "localhost:ipp"
    )

    # We can't match against endpoint address "*"
    assert (
        lsof_to_file(
            ["p123", "f6", "au", "tIPv6", "d0x42", "n*:57919"]
        ).get_endpoints()[0]
        is None
    )
    assert (
        lsof_to_file(
            ["p123", "f6", "au", "tIPv4", "d0x42", "n*:57919"]
        ).get_endpoints()[0]
        is None
    )

    assert (
        lsof_to_file(["p123", "f6", "au", "tIPv4", "d0x42", "n*:*"]).get_endpoints()[0]
        is None
    )
    assert (
        lsof_to_file(
            ["p123", "f6", "aw", "tREG", "d0x42", "n/somefile"]
        ).get_endpoints()[0]
        is None
    )


def test_remote_endpoint():
    remote_endpoint = lsof_to_file(
        ["p123", "f6", "au", "tIPv4", "d0x42", "nlocalhost:postgresql->localhost:3331"]
    ).get_endpoints()[1]
    assert remote_endpoint == "localhost:3331"

    remote_endpoint = lsof_to_file(
        ["p123", "f6", "au", "tIPv4", "d0x42", "nlocalhost:postgresql->otherhost:3331"]
    ).get_endpoints()[1]
    assert remote_endpoint == "otherhost:3331"

    assert (
        lsof_to_file(
            ["p123", "f6", "au", "tIPv6", "d0x42", "nlocalhost:19091"]
        ).get_endpoints()[1]
        is None
    )
    assert (
        lsof_to_file(
            ["p123", "f6", "au", "tIPv6", "d0x42", "n*:57919"]
        ).get_endpoints()[1]
        is None
    )
    assert (
        lsof_to_file(
            ["p123", "f6", "au", "tIPv4", "d0x42", "n*:57919"]
        ).get_endpoints()[1]
        is None
    )

    assert (
        lsof_to_file(["p123", "f6", "au", "tIPv4", "d0x42", "n*:*"]).get_endpoints()[1]
        is None
    )
    assert (
        lsof_to_file(
            ["p123", "f6", "aw", "tREG", "d0x42", "n/somefile"]
        ).get_endpoints()[1]
        is None
    )


def test_str_resolve():
    # FIXME: This will break if Google changes the name of 8.8.8.8
    test_me = px_file.PxFile(pid=0, filetype="IPv4")
    test_me.name = "127.0.0.1:51786->8.8.8.8:https"
    assert str(test_me) in [
        "[IPv4] localhost:51786->google-public-dns-a.google.com:https",
        "[IPv4] localhost:51786->dns.google:https",
    ]

    test_me = px_file.PxFile(pid=0, filetype="IPv4")
    test_me.name = "127.0.0.1:17600"
    assert str(test_me) == "[IPv4] localhost:17600 (LISTEN)"

    test_me = px_file.PxFile(pid=0, filetype="IPv6")
    test_me.name = "[::1]:17600"

    match = re.match(r"^\[IPv6\] (.*):17600 \(LISTEN\)$", str(test_me))
    assert match
    resolution = match.group(1)
    assert resolution == "[::1]" or "localhost" in resolution

    test_me = px_file.PxFile(pid=0, filetype="IPv4")
    test_me.name = "this:is:garbage:17600"
    assert str(test_me) == "[IPv4] this:is:garbage:17600 (LISTEN)"



#!/usr/bin/env python3

"""
Build the project from a clean clone to make sure that works.
"""

import os
import shutil
import subprocess
import sys
import tempfile

# Copy everything to a temporary directory
tempdir = tempfile.mkdtemp()
print(f"Copying sources into <{tempdir}>...")
shutil.copytree(
    ".",
    tempdir,
    dirs_exist_ok=True,
    ignore=shutil.ignore_patterns(
        ".tox", "env", ".mypy_cache", ".pytest_cache", "__pycache__"
    ),
)
os.chdir(tempdir)

# Remove any artifacts so the build is clean
print("Cleaning sources...")
subprocess.run(
    [
        "git",
        "clean",
        "-d",
        "--force",
        "-x",
        "--quiet",
    ],
    check=True,
)

# Build the clone
print("Building sources using python -m build...")
FAKE_VERSION = "99.99.99"
subprocess.run(["git", "config", "user.email", "you@example.com"], check=True)
subprocess.run(["git", "config", "user.name", "Your Name"], check=True)
subprocess.run(
    [
        "git",
        "commit",
        "--all",
        "--allow-empty",
        "--message",
        "Ensure we're clean before building the wheel",
    ],
    check=True,
)
subprocess.run(
    [
        "git",
        "tag",
        "--annotate",
        "--force",
        "--message",
        "Forced tag for testing the build",
        FAKE_VERSION,
    ],
    check=True,
)

# This is what tox.ini does in test-wheel
subprocess.run(
    [
        "python3",
        "-m",
        "build",
        "--outdir",
        "dist",
    ],
    check=True,
)

# Check that no files in dist have "0.0.0" in their names
print(f"Checking that all files in dist have '{FAKE_VERSION}' in their names...")
found_misversioned = False
for filename in os.listdir("dist"):
    if FAKE_VERSION not in filename:
        print(f"{FAKE_VERSION} not in <{filename}>")
        found_misversioned = True
if found_misversioned:
    print(f"{FAKE_VERSION} not found all dist/ filenames, failing...")
    sys.exit(1)

shutil.rmtree(tempdir)



import subprocess

from px import px_exec_util


def test_exec_false_no_check():
    assert px_exec_util.run(["false"]) == ""


def test_exec_true_with_check():
    assert px_exec_util.run(["true"], check_exitcode=True) == ""


def test_exec_false_with_check():
    try:
        px_exec_util.run(["false"], check_exitcode=True)
        assert False and "We should never get here"
    except subprocess.CalledProcessError:
        # This is the exception we want, done!
        pass



import os
import re

from px import px_commandline


def test_get_coalesce_candidate():
    assert px_commandline.get_trailing_absolute_path("/hello") == "/hello"
    assert px_commandline.get_trailing_absolute_path("/hello:/baloo") == "/baloo"

    assert px_commandline.get_trailing_absolute_path("-Dx=/hello") == "/hello"
    assert px_commandline.get_trailing_absolute_path("-Dx=/hello:/baloo") == "/baloo"

    assert px_commandline.get_trailing_absolute_path("hello") is None
    assert px_commandline.get_trailing_absolute_path("hello:/baloo") is None

    assert (
        px_commandline.get_trailing_absolute_path(
            "/A/IntelliJ IDEA.app/C/p/mm/lib/mm.jar:/A/IntelliJ IDEA.app/C/p/ms/lib/ms.jar:/A/IntelliJ"
        )
        == "/A/IntelliJ"
    )


def test_coalesce_count():
    def exists(s):
        return s in ["/", "/a b c", "/a b c/"]

    assert px_commandline.coalesce_count(["/a", "b", "c"], exists=exists) == 3
    assert px_commandline.coalesce_count(["/a", "b", "c/"], exists=exists) == 3
    assert px_commandline.coalesce_count(["/a", "b", "c", "d"], exists=exists) == 3

    assert (
        px_commandline.coalesce_count(["/a", "b", "c:/a", "b", "c"], exists=exists) == 5
    )
    assert (
        px_commandline.coalesce_count(["/a", "b", "c/:/a", "b", "c/"], exists=exists)
        == 5
    )

    assert (
        px_commandline.coalesce_count(["/a", "b", "c:/a", "b", "c", "d"], exists=exists)
        == 5
    )
    assert (
        px_commandline.coalesce_count(
            ["/a", "b", "c/:/a", "b", "c/", "d/"], exists=exists
        )
        == 5
    )


def test_to_array_spaced1():
    assert px_commandline.to_array(
        "java -Dhello=/Applications/IntelliJ IDEA.app/Contents",
        exists=lambda s: s
        in [
            "/Applications",
            "/Applications/IntelliJ IDEA.app",
            "/Applications/IntelliJ IDEA.app/Contents",
        ],
    ) == ["java", "-Dhello=/Applications/IntelliJ IDEA.app/Contents"]


def test_to_array_spaced2():
    assert px_commandline.to_array(
        " ".join(
            [
                "java",
                "-Dhello=/Applications/IntelliJ IDEA.app/Contents/Info.plist",
                "-classpath",
                "/Applications/IntelliJ",
                "IDEA.app/Contents/plugins/maven-model/lib/maven-model.jar:/Applications/IntelliJ",
                "IDEA.app/Contents/plugins/maven-server/lib/maven-server.jar:/Applications/IntelliJ",
                "IDEA.app/Contents/plugins/maven/lib/maven3-server-common.jar",
                "MainClass",
            ]
        ),
        exists=lambda s: s
        in [
            "/Applications",
            "/Applications/IntelliJ IDEA.app/Contents/Info.plist",
            "/Applications/IntelliJ IDEA.app/Contents/plugins/maven-model/lib/maven-model.jar",
            "/Applications/IntelliJ IDEA.app/Contents/plugins/maven-server/lib/maven-server.jar",
            "/Applications/IntelliJ IDEA.app/Contents/plugins/maven/lib/maven3-server-common.jar",
        ],
    ) == [
        "java",
        "-Dhello=/Applications/IntelliJ IDEA.app/Contents/Info.plist",
        "-classpath",
        "/Applications/IntelliJ IDEA.app/Contents/plugins/maven-model/lib/maven-model.jar:/Applications/IntelliJ IDEA.app/Contents/plugins/maven-server/lib/maven-server.jar:/Applications/IntelliJ IDEA.app/Contents/plugins/maven/lib/maven3-server-common.jar",
        "MainClass",
    ]


def test_to_array_spaced3():
    """Same as test_to_array_spaced2() but with two spaces rather than just one."""
    assert px_commandline.to_array(
        " ".join(
            [
                "java",
                "-Dhello=/Applications/IntelliJ IDEA CE.app/Contents/Info.plist",
                "-classpath",
                "/Applications/IntelliJ",
                "IDEA",
                "CE.app/Contents/plugins/maven-model/lib/maven-model.jar:/Applications/IntelliJ",
                "IDEA",
                "CE.app/Contents/plugins/maven-server/lib/maven-server.jar:/Applications/IntelliJ",
                "IDEA",
                "CE.app/Contents/plugins/maven/lib/maven3-server-common.jar",
                "MainClass",
            ]
        ),
        exists=lambda s: s
        in [
            "/Applications",
            "/Applications/IntelliJ IDEA CE.app/Contents/Info.plist",
            "/Applications/IntelliJ IDEA CE.app/Contents/plugins/maven-model/lib/maven-model.jar",
            "/Applications/IntelliJ IDEA CE.app/Contents/plugins/maven-server/lib/maven-server.jar",
            "/Applications/IntelliJ IDEA CE.app/Contents/plugins/maven/lib/maven3-server-common.jar",
        ],
    ) == [
        "java",
        "-Dhello=/Applications/IntelliJ IDEA CE.app/Contents/Info.plist",
        "-classpath",
        "/Applications/IntelliJ IDEA CE.app/Contents/plugins/maven-model/lib/maven-model.jar:/Applications/IntelliJ IDEA CE.app/Contents/plugins/maven-server/lib/maven-server.jar:/Applications/IntelliJ IDEA CE.app/Contents/plugins/maven/lib/maven3-server-common.jar",
        "MainClass",
    ]


def test_to_array_ms_edge():
    complete = "/".join(
        [
            "/Applications",
            "Microsoft Edge.app",
            "Contents",
            "Frameworks",
            "Microsoft Edge Framework.framework",
            "Versions",
            "122.0.2365.63",
            "Helpers",
            "Microsoft Edge Helper (GPU).app",
            "Contents",
            "MacOS",
            "Microsoft Edge Helper (GPU)",
        ]
    )
    exists = []
    partial = complete
    while True:
        exists.append(partial)
        partial = os.path.dirname(partial)
        if partial == "/":
            break

    assert px_commandline.to_array(
        complete + " --type=gpu-process",
        exists=lambda s: s in exists,
    ) == [complete] + ["--type=gpu-process"]


def test_get_command_python():
    assert px_commandline.get_command("python") == "python"
    assert px_commandline.get_command("/apa/Python") == "Python"
    assert px_commandline.get_command("python --help") == "python"

    # These are inspired by Python 2.7.11 --help output
    assert px_commandline.get_command("python apa.py") == "apa.py"
    assert px_commandline.get_command("python /usr/bin/apa.py") == "apa.py"
    assert px_commandline.get_command("python2.7 /usr/bin/apa.py") == "apa.py"
    assert px_commandline.get_command("python /usr/bin/hej") == "hej"
    assert px_commandline.get_command("python /usr/bin/hej gris --flaska") == "hej"
    assert px_commandline.get_command("python -c cmd") == "python"
    assert px_commandline.get_command("python -m mod") == "mod"
    assert px_commandline.get_command("python -m mod --hej gris --frukt") == "mod"
    assert px_commandline.get_command("Python -") == "Python"

    # Ignoring switches
    assert px_commandline.get_command("python -E apa.py") == "apa.py"
    assert px_commandline.get_command("python3 -E") == "python3"
    assert px_commandline.get_command("python -u -t -m mod") == "mod"

    # -W switches unsupported for now, room for future improvement
    assert px_commandline.get_command("python -W warning:spec apa.py") == "python"

    # Invalid command lines
    assert px_commandline.get_command("python -W") == "python"
    assert px_commandline.get_command("python -c") == "python"
    assert px_commandline.get_command("python -m") == "python"
    assert px_commandline.get_command("python -m   ") == "python"
    assert px_commandline.get_command("python -m -u") == "python"
    assert px_commandline.get_command("python    ") == "python"


def test_get_command_guile():
    assert px_commandline.get_command("guile") == "guile"
    assert px_commandline.get_command("guile --help") == "guile"

    assert px_commandline.get_command("guile apa.scm") == "apa.scm"
    assert px_commandline.get_command("guile /usr/bin/apa.scm") == "apa.scm"
    assert px_commandline.get_command("guile /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile -L foo /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile -L foo -C bar /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile -x ex /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile -l bar /usr/bin/hej") == "bar"
    assert px_commandline.get_command("guile -l bar") == "bar"
    assert px_commandline.get_command("guile -l") == "guile"
    assert px_commandline.get_command("guile -e quux /usr/bin/hej") == "hej"
    assert (
        px_commandline.get_command("guile --language=ecmascript /usr/bin/hej") == "hej"
    )
    assert px_commandline.get_command("guile --debug /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile --no-debug /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile --auto-compile /usr/bin/hej") == "hej"
    assert (
        px_commandline.get_command("guile --fresh-auto-compile /usr/bin/hej") == "hej"
    )
    assert px_commandline.get_command("guile --no-auto-compile /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile --listen /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile -q /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile --use-srfi=2,13,14 /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile --r6rs /usr/bin/hej") == "hej"
    assert px_commandline.get_command("guile --r7rs /usr/bin/hej") == "hej"


def test_get_command_aws():
    assert px_commandline.get_command("Python /usr/local/bin/aws") == "aws"
    assert px_commandline.get_command("python aws s3") == "aws s3"
    assert px_commandline.get_command("python3 aws s3 help") == "aws s3 help"
    assert (
        px_commandline.get_command("/wherever/python3 aws s3 help flaska")
        == "aws s3 help"
    )

    assert px_commandline.get_command("python aws s3 sync help") == "aws s3 sync help"
    assert px_commandline.get_command("python aws s3 sync nothelp") == "aws s3 sync"

    assert px_commandline.get_command("python aws s3 --unknown sync") == "aws s3"

    assert (
        px_commandline.get_command(
            " ".join(
                [
                    "python3",
                    "/usr/local/bin/aws",
                    "--profile=system-admin-prod",
                    "--region=eu-west-1",
                    "s3",
                    "sync",
                    "--only-show-errors",
                    "s3://xxxxxx",
                    "./xxxxxx",
                ]
            )
        )
        == "aws s3 sync"
    )


def test_get_command_java():
    assert px_commandline.get_command("java") == "java"
    assert px_commandline.get_command("java -version") == "java"
    assert px_commandline.get_command("java -help") == "java"

    assert px_commandline.get_command("java SomeClass") == "SomeClass"
    assert px_commandline.get_command("java x.y.SomeClass") == "SomeClass"
    assert px_commandline.get_command("java -jar flaska.jar") == "flaska.jar"
    assert px_commandline.get_command("java -jar /a/b/flaska.jar") == "flaska.jar"

    # Special handling of class name "Main"
    assert px_commandline.get_command("java a.b.c.Main") == "c.Main"

    # We should ignore certain command line parameters
    assert px_commandline.get_command("java -server SomeClass") == "SomeClass"
    assert px_commandline.get_command("java -Xwhatever SomeClass") == "SomeClass"
    assert px_commandline.get_command("java -Dwhatever SomeClass") == "SomeClass"
    assert px_commandline.get_command("java -cp /a/b/c SomeClass") == "SomeClass"
    assert px_commandline.get_command("java -classpath /a/b/c SomeClass") == "SomeClass"
    assert px_commandline.get_command("java -eahej SomeClass") == "SomeClass"
    assert px_commandline.get_command("java -dahej SomeClass") == "SomeClass"

    # Tests for invalid command lines
    assert px_commandline.get_command("java -cp /a/b/c") == "java"
    assert px_commandline.get_command("java  ") == "java"
    assert px_commandline.get_command("java -jar") == "java"
    assert px_commandline.get_command("java -jar    ") == "java"


def test_get_command_java_gradled():
    commandline = (
        "/Library/Java/JavaVirtualMachines/jdk1.8.0_60.jdk/Contents/Home/bin/java "
        + "-XX:MaxPermSize=256m -XX:+HeapDumpOnOutOfMemoryError -Xmx1024m "
        + "-Dfile.encoding=UTF-8 -Duser.country=SE -Duser.language=sv -Duser.variant -cp "
        + "/Users/johan/.gradle/wrapper/dists/gradle-2.8-all/gradle-2.8/lib/gradle-launcher-2.8.jar"
        + " org.gradle.launcher.daemon.bootstrap.GradleDaemon 2.8"
    )
    assert px_commandline.get_command(commandline) == "GradleDaemon"


def test_get_command_java_teamcity():
    commandline = (
        "/usr/lib/jvm/jdk-8-oracle-x64/jre/bin/java "
        + "-Djava.util.logging.config.file=/teamcity/conf/logging.properties "
        + "-Djava.util.logging.manager=org.apache.juli.ClassLoaderLogManager "
        + "-Dsun.net.inetaddr.ttl=60 -server -Xms31g -Xmx31g "
        + "-Dteamcity.configuration.path=../conf/teamcity-startup.properties "
        + "-Dlog4j.configuration=file:/teamcity/bin/../conf/teamcity-server-log4j.xml "
        + "-Dteamcity_logs=../logs/ -Djsse.enableSNIExtension=false -Djava.awt.headless=true "
        + "-Djava.endorsed.dirs=/teamcity/endorsed "
        + "-classpath /teamcity/bin/bootstrap.jar:/teamcity/bin/tomcat-juli.jar "
        + "-Dcatalina.base=/teamcity -Dcatalina.home=/teamcity "
        + "-Djava.io.tmpdir=/teamcity/temp org.apache.catalina.startup.Bootstrap start"
    )
    assert px_commandline.get_command(commandline) == "Bootstrap"


def test_get_command_java_logstash():
    # From: https://github.com/elastic/logstash/issues/3315
    commandline = (
        "/usr/bin/java -XX:+UseParNewGC -XX:+UseConcMarkSweepGC "
        + "-Djava.awt.headless=true -XX:CMSInitiatingOccupancyFraction=75 "
        + "-XX:+UseCMSInitiatingOccupancyOnly -Djava.io.tmpdir=/var/lib/logstash "
        + "-Xmx128m -Xss2048k -Djffi.boot.library.path=/opt/logstash/vendor/jruby/lib/jni "
        + "-XX:+UseParNewGC -XX:+UseConcMarkSweepGC -Djava.awt.headless=true "
        + "-XX:CMSInitiatingOccupancyFraction=75 -XX:+UseCMSInitiatingOccupancyOnly "
        + "-Djava.io.tmpdir=/var/lib/logstash "
        + "-Xbootclasspath/a:/opt/logstash/vendor/jruby/lib/jruby.jar -classpath : "
        + "-Djruby.home=/opt/logstash/vendor/jruby "
        + "-Djruby.lib=/opt/logstash/vendor/jruby/lib -Djruby.script=jruby "
        + "-Djruby.shell=/bin/sh org.jruby.Main --1.9 /opt/logstash/lib/bootstrap/environment.rb "
        + "logstash/runner.rb agent -f /etc/logstash/conf.d -l /var/log/logstash/logstash.log"
    )
    assert px_commandline.get_command(commandline) == "jruby.Main"


def test_get_command_java_gradleworkermain():
    commandline = (
        "/some/path/bin/java "
        "-Djava.awt.headless=true "
        "-Djava.security.manager="
        + "worker.org.gradle.process.internal.worker.child.BootstrapSecurityManager "
        "-Dorg.gradle.native=false "
        "-Drobolectric.accessibility.enablechecks=true "
        "-Drobolectric.logging=stderr "
        "-Drobolectric.logging.enabled=true "
        "-agentlib:jdwp=transport=dt_socket,server=y,address=,suspend=n "
        "-noverify "
        "-javaagent:gen_build/.../jacocoagent.jar="
        + "destfile=gen_build/jacoco/testDebugUnitTest.exec,"
        + "append=true,dumponexit=true,output=file,jmx=false "
        "-Xmx2400m "
        "-Dfile.encoding=UTF-8 "
        "-Duser.country=SE "
        "-Duser.language=sv "
        "-Duser.variant "
        "-ea "
        "-cp "
        "/Users/walles/.gradle/caches/4.2.1/workerMain/gradle-worker.jar "
        "worker.org.gradle.process.internal.worker.GradleWorkerMain "
        "'Gradle Test Executor 16'"
    )
    assert px_commandline.get_command(commandline) == "GradleWorkerMain"


def test_get_command_java_equinox():
    # You get these processes with Visual Studio Code's java environment in October 2020
    commandline = (
        "/Library/Java/JavaVirtualMachines/openjdk-11.jdk/Contents/Home/bin/java "
        "--add-modules=ALL-SYSTEM "
        "--add-opens "
        "java.base/java.util=ALL-UNNAMED "
        "--add-opens "
        "java.base/java.lang=ALL-UNNAMED "
        "-Declipse.application=org.eclipse.jdt.ls.core.id1 "
        "-Dosgi.bundles.defaultStartLevel=4 "
        "-Declipse.product=org.eclipse.jdt.ls.core.product "
        "-Dfile.encoding=utf8 "
        "-XX:+UseParallelGC "
        "-XX:GCTimeRatio=4 "
        "-XX:AdaptiveSizePolicyWeight=90 "
        "-Dsun.zip.disableMemoryMapping=true "
        "-Xmx1G "
        "-Xms100m "
        "-noverify "
        "-jar "
        "/Users/walles/.vscode/extensions/redhat.java-0.68.0/server/plugins/org.eclipse.equinox.launcher_1.5.800.v20200727-1323.jar "
        "-configuration "
        "/Users/walles/Library/Application "
        "Support/Code/User/globalStorage/redhat.java/0.68.0/config_mac "
        "-data "
        "/Users/walles/Library/Application "
        "Support/Code/User/workspaceStorage/b8c3a38f62ce0fc92ce4edfb836480db/redhat.java/jdt_ws "
    )
    assert (
        px_commandline.get_command(commandline)
        == "org.eclipse.equinox.launcher_1.5.800.v20200727-1323.jar"
    )


# Ref: https://github.com/walles/px/issues/137
def test_get_command_line_jdk_java_options():
    commandline = """
/opt/homebrew/Cellar/openjdk/24.0.1/libexec/openjdk.jdk/Contents/Home/bin/java
  -classpath
  /opt/homebrew/Cellar/mvnd/2.0.0-rc-3/libexec/mvn/boot/plexus-classworlds-2.8.0.jar
  -javaagent:/opt/homebrew/Cellar/mvnd/2.0.0-rc-3/libexec/mvn/lib/mvnd/mvnd-agent-2.0.0-rc-3.jar
  -Dmvnd.home=/opt/homebrew/Cellar/mvnd/2.0.0-rc-3/libexec
  -Dmaven.home=/opt/homebrew/Cellar/mvnd/2.0.0-rc-3/libexec/mvn
  -Dmaven.conf=/opt/homebrew/Cellar/mvnd/2.0.0-rc-3/libexec/mvn/conf
  -Dclassworlds.conf=/opt/homebrew/Cellar/mvnd/2.0.0-rc-3/libexec/bin/mvnd-daemon.conf
  -Dmaven.logger.logFile=/Users/johan/.m2/mvnd/registry/2.0.0-rc-3/daemon-802431a9.log
  -Dmvnd.java.home=/opt/homebrew/Cellar/openjdk/24.0.1/libexec/openjdk.jdk/Contents/Home
  -Dmvnd.id=802431a9
  -Dmvnd.daemonStorage=/Users/johan/.m2/mvnd/registry/2.0.0-rc-3
  -Dmvnd.registry=/Users/johan/.m2/mvnd/registry/2.0.0-rc-3/registry.bin
  -Dmvnd.socketFamily=inet
  -Dmvnd.home=/opt/homebrew/Cellar/mvnd/2.0.0-rc-3/libexec
  -Djdk.java.options=--add-opens
  java.base/java.io=ALL-UNNAMED
  --add-opens
  java.base/java.lang=ALL-UNNAMED
  --add-opens
  java.base/java.util=ALL-UNNAMED
  --add-opens
  java.base/jdk.internal.misc=ALL-UNNAMED
  --add-opens
  java.base/sun.net.www.protocol.jar=ALL-UNNAMED
  --add-opens
  java.base/sun.nio.fs=ALL-UNNAMED
  -Dmvnd.noDaemon=false
  -Dmvnd.debug=false
  -Dmvnd.debug.address=8000
  -Dmvnd.idleTimeout=3h
  -Dmvnd.keepAlive=100ms
  -Dmvnd.extClasspath=
  -Dmvnd.coreExtensionsDiscriminator=da39a3ee5e6b4b0d3255bfef95601890afd80709
  -Dmvnd.coreExtensionsExclude=io.takari.maven:takari-smart-builder
  -Dmvnd.enableAssertions=false
  -Dmvnd.expirationCheckDelay=10s
  -Dmvnd.duplicateDaemonGracePeriod=10s
  -Dmvnd.socketFamily=inet
  org.codehaus.plexus.classworlds.launcher.Launcher
""".strip().replace("\n", " ")
    commandline = re.sub(r"\s+", " ", commandline)

    assert px_commandline.get_command(commandline) == "Launcher"


# Ref: https://github.com/walles/px/issues/139
def test_get_command_line_java_issue_139():
    commandline = """
/opt/homebrew/Cellar/openjdk@21/21.0.8/libexec/openjdk.jdk/Contents/Home/bin/java
  --add-exports=jdk.compiler/com.sun.tools.javac.api=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.file=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.main=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.model=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.parser=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.processing=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.tree=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.util=ALL-UNNAMED
  --add-opens=jdk.compiler/com.sun.tools.javac.code=ALL-UNNAMED
  --add-opens=jdk.compiler/com.sun.tools.javac.comp=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.code=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.api=ALL-UNNAMED
  --add-exports=jdk.compiler/com.sun.tools.javac.util=ALL-UNNAMED
  @/Users/johan/.gradle/.tmp/gradle-worker-classpath12697647132064750531txt
  -Xmx2g
  -Dfile.encoding=UTF-8
  -Duser.country=SE
  -Duser.language=sv
  -Duser.variant
  worker.org.gradle.process.internal.worker.GradleWorkerMain
  'Gradle
  Worker
  Daemon
  3'
""".strip().replace("\n", " ")
    commandline = re.sub(r"\s+", " ", commandline)

    assert px_commandline.get_command(commandline) == "GradleWorkerMain"


def test_prettify_java_class():
    assert (
        px_commandline.prettify_fully_qualified_java_class("com.example.MyClass")
        == "MyClass"
    )
    assert (
        px_commandline.prettify_fully_qualified_java_class("com.example.Main")
        == "example.Main"
    )
    assert px_commandline.prettify_fully_qualified_java_class("") is None


# Ref: https://github.com/walles/px/issues/126
def test_get_command_flutter():
    # Candidate contains no slashes => subcommand
    assert (
        px_commandline.get_command(
            "/usr/local/Caskroom/flutter/3.0.1/flutter/bin/cache/dart-sdk/bin/dart "
            "devtools "
            "--machine "
            "--allow-embedding "
        )
        == "dart devtools"
    )

    # Candidate contains slashes => file to run
    assert (
        px_commandline.get_command(
            "/usr/local/Caskroom/flutter/3.0.1/flutter/bin/cache/dart-sdk/bin/dart "
            "--disable-dart-dev "
            "--packages=/usr/local/Caskroom/flutter/3.0.1/flutter/packages/flutter_tools/.dart_tool/package_config.json "
            "/usr/local/Caskroom/flutter/3.0.1/flutter/bin/cache/flutter_tools "
            "upgrade"
        )
        == "flutter_tools"
    )

    # Candidate contains dots (.) => file to run
    assert (
        px_commandline.get_command(
            "/usr/local/Caskroom/flutter/3.0.1/flutter/bin/cache/dart-sdk/bin/dart "
            "--verbosity=error "
            "--disable-dart-dev "
            "--snapshot=/usr/local/Caskroom/flutter/3.0.1/flutter/bin/cache/flutter_tools.snapshot "
            "--snapshot-kind=app-jit "
            "--packages=/usr/local/Caskroom/flutter/3.0.1/flutter/packages/flutter_tools/.dart_tool/package_config.json "
            "--no-enable-mirrors "
            "flutter_tools.dart "
        )
        == "flutter_tools.dart"
    )


def test_get_command_electron_macos():
    # Note that if we have spaces inside of the path, the path in
    # question needs to be valid on the local system for it to
    # be deciphered properly by px. So this path is from an actual
    # path, but with all spaces removed.
    assert (
        px_commandline.get_command(
            "/Applications/VisualStudioCode.app/Contents/MacOS/Electron "
            "--ms-enable-electron-run-as-node "
            "/Users/johan/.vscode/extensions/ms-python.vscode-pylance-2021.12.2/dist/server.bundle.js "
            "--cancellationReceive=file:d6fe53594ec46a8bb986ad058c985f56d309e7bf19 "
            "--node-ipc "
            "--clientProcessId=42516 "
        )
        == "VisualStudioCode"
    )


def test_get_command_resque():
    # These command names are from a real-world system
    assert px_commandline.get_command("resque-1.20.0: x y z") == "resque-1.20.0:"
    assert px_commandline.get_command("resqued-0.7.12 a b c") == "resqued-0.7.12"


def test_get_command_sudo():
    assert px_commandline.get_command("sudo") == "sudo"
    assert (
        px_commandline.get_command("sudo python /usr/bin/hej gris --flaska")
        == "sudo hej"
    )

    # We could support flags at some point, but until proven necessary we'll
    # fall back on just "sudo" when we get flags.
    assert px_commandline.get_command("sudo -B python /usr/bin/hej") == "sudo"


def test_get_command_sudo_with_space_in_command_name(tmpdir):
    # Create a file name with a space in it
    spaced_path = tmpdir.join("i contain spaces")
    spaced_path.write_binary(b"")
    spaced_name = str(spaced_path)

    # Verify splitting of the spaced file name
    assert px_commandline.get_command("sudo " + spaced_name) == "sudo i contain spaces"

    # Verify splitting with more parameters on the line
    assert (
        px_commandline.get_command("sudo " + spaced_name + " parameter")
        == "sudo i contain spaces"
    )


def test_get_command_sudo_with_space_in_path(tmpdir):
    # Create a file name with a space in it
    spaced_dir = tmpdir.join("i contain spaces")
    os.mkdir(str(spaced_dir))
    spaced_path = spaced_dir + "/runme"

    spaced_path.write_binary(b"")
    spaced_name = str(spaced_path)

    # Verify splitting of the spaced file name
    assert px_commandline.get_command("sudo " + spaced_name) == "sudo runme"

    # Verify splitting with more parameters on the line
    assert (
        px_commandline.get_command("sudo " + spaced_name + " parameter") == "sudo runme"
    )


def test_get_command_interpreters():
    assert px_commandline.get_command("ruby") == "ruby"
    assert px_commandline.get_command("ruby /some/path/apa.rb") == "apa.rb"
    assert px_commandline.get_command("ruby -option /some/path/apa.rb") == "ruby"

    assert px_commandline.get_command("sh") == "sh"
    assert px_commandline.get_command("sh /some/path/apa.sh") == "apa.sh"
    assert px_commandline.get_command("sh -option /some/path/apa.sh") == "sh"

    assert px_commandline.get_command("bash") == "bash"
    assert px_commandline.get_command("bash /some/path/apa.sh") == "apa.sh"
    assert px_commandline.get_command("bash -option /some/path/apa.sh") == "bash"

    assert px_commandline.get_command("perl") == "perl"
    assert px_commandline.get_command("perl /some/path/apa.pl") == "apa.pl"
    assert px_commandline.get_command("perl -option /some/path/apa.pl") == "perl"


def test_get_command_unicode():
    assert px_commandline.get_command("😀") == "😀"


def test_get_command_ruby_switches():
    assert (
        px_commandline.get_command(
            "/usr/bin/ruby -W0 /usr/local/bin/brew.rb install rust"
        )
        == "brew.rb install"
    )

    # https://github.com/walles/px/issues/87
    assert (
        px_commandline.get_command("/usr/bin/ruby -W1 -- /apa/build.rb /bepa/cmake.rb")
        == "build.rb"
    )

    # https://github.com/walles/px/issues/74
    assert (
        px_commandline.get_command(
            "/usr/bin/ruby -Eascii-8bit:ascii-8bit /usr/sbin/google-fluentd"
        )
        == "google-fluentd"
    )


def test_get_command_perl():
    # Source: https://github.com/walles/px/issues/85
    assert (
        px_commandline.get_command(
            " ".join(
                [
                    "/usr/bin/perl5.18",
                    "/usr/local/Cellar/cloc/1.90/libexec/bin/cloc",
                    "build-system",
                    "build_number_offset",
                    "buildbox",
                    "Random.txt",
                    "README.md",
                    "submodules",
                    "Telegram",
                    "third-party",
                    "tools",
                    "versions.json",
                    "WORKSPACE",
                ]
            )
        )
        == "cloc"
    )

    # Variations on the same theme
    assert (
        px_commandline.get_command(
            " ".join(
                [
                    "/usr/bin/perl",
                    "/usr/local/Cellar/cloc/1.90/libexec/bin/cloc",
                ]
            )
        )
        == "cloc"
    )
    assert (
        px_commandline.get_command(
            " ".join(
                [
                    "perl",
                    "/usr/local/Cellar/cloc/1.90/libexec/bin/cloc",
                ]
            )
        )
        == "cloc"
    )
    assert (
        px_commandline.get_command(
            " ".join(
                [
                    "/usr/bin/perl5",
                    "/usr/local/Cellar/cloc/1.90/libexec/bin/cloc",
                ]
            )
        )
        == "cloc"
    )
    assert (
        px_commandline.get_command(
            " ".join(
                [
                    "/usr/bin/perl5.30",
                    "/usr/local/Cellar/cloc/1.90/libexec/bin/cloc",
                ]
            )
        )
        == "cloc"
    )

    # Give up on command line switches
    assert (
        px_commandline.get_command(
            " ".join(
                [
                    "/usr/bin/perl",
                    "-S",
                    "cloc",
                ]
            )
        )
        == "perl"
    )


def test_get_homebrew_commandline():
    # Source: https://github.com/walles/px/issues/72
    assert (
        px_commandline.get_command(
            " ".join(
                [
                    "/usr/local/Homebrew/Library/Homebrew/vendor/portable-ruby/current/bin/ruby",
                    "-W0",
                    "--disable=gems,did_you_mean,rubyopt",
                    "/usr/local/Homebrew/Library/Homebrew/brew.rb",
                    "upgrade",
                ]
            )
        )
        == "brew.rb upgrade"
    )


def test_get_bash_brew_sh_commandline():
    assert (
        px_commandline.get_command(
            "/bin/bash -p /usr/local/Homebrew/Library/Homebrew/brew.sh upgrade"
        )
        == "brew.sh upgrade"
    )


def test_get_terraform_provider_commandline():
    # Source: https://github.com/walles/px/issues/105
    assert (
        px_commandline.get_command(
            ".terraform/providers/registry.terraform.io/heroku/heroku/4.8.0/darwin_amd64/terraform-provider-heroku_v4.8.0"
        )
        == "terraform-provider-heroku_v4.8.0"
    )


def test_get_terraform_commandline():
    # Source: https://github.com/walles/px/issues/113
    assert (
        px_commandline.get_command("terraform -chdir=dev apply -target=abc123")
        == "terraform apply"
    )


def test_get_go_commandline():
    assert px_commandline.get_command("go build ./...") == "go build"
    assert px_commandline.get_command("go --version") == "go"
    assert px_commandline.get_command("/usr/local/bin/go") == "go"


def test_get_git_commandline():
    assert (
        px_commandline.get_command("git clone git@github.com:walles/riff")
        == "git clone"
    )
    assert px_commandline.get_command("git --version") == "git"
    assert px_commandline.get_command("/usr/local/bin/git") == "git"


def test_node_max_old_space():
    assert (
        px_commandline.get_command("node --max_old_space_size=4096 scripts/start.js")
        == "start.js"
    )


def test_dotnet_commandline():
    # Unclear whether "fable" is a builtin or a separate tool, go with "dotnet fable".
    assert (
        px_commandline.get_command("libexec/dotnet fable core/Core.Test.fsproj")
        == "dotnet fable"
    )

    # The DLL has a path, so it can't be a builtin. Go with just "fable.dll"
    assert (
        px_commandline.get_command("libexec/dotnet any/fable.dll core/Core.Test.fsproj")
        == "fable.dll"
    )


def test_macos_app():
    assert (
        px_commandline.get_command(
            "/".join(
                [
                    "/System",
                    "Library",
                    "CoreServices",
                    "Dock.app",
                    "Contents",
                    "XPCServices",
                    "com.apple.dock.external.extra.xpc",
                    "Contents",
                    "MacOS",
                    "com.apple.dock.external.extra",
                ]
            )
        )
        == "Dock/extra"
    )

    # https://github.com/walles/px/issues/73
    assert (
        px_commandline.get_command(
            "/".join(
                [
                    "/Applications",
                    "Firefox.app",
                    "Contents",
                    "MacOS",
                    "plugin-container.app",
                    "Contents",
                    "MacOS",
                    "plugin-container",
                ]
            )
        )
        == "Firefox/plugin-container"
    )

    # Note that if we have spaces inside of the path, the path in
    # question needs to be valid on the local system for it to
    # be deciphered properly by px. So this path is from an actual
    # path, but with all spaces removed.
    assert (
        px_commandline.get_command(
            "/".join(
                [
                    "/Applications",
                    "VisualStudioCode.app",
                    "Contents",
                    "Frameworks",
                    "CodeHelper(Renderer).app",
                    "Contents",
                    "MacOS",
                    "CodeHelper(Renderer)",
                ]
            )
        )
        == "CodeHelper(Renderer)"
    )

    assert (
        px_commandline.get_command("/Applications/iTerm.app/Contents/MacOS/iTerm2")
        == "iTerm2"
    )

    # Don't duplicate the .app name
    assert (
        px_commandline.get_command(
            "/".join(
                [
                    "/System",
                    "Library",
                    "PrivateFrameworks",
                    "IDS.framework",
                    "identityservicesd.app",
                    "Contents",
                    "MacOS",
                    "identityservicesd",
                ]
            )
        )
        == "IDS/identityservicesd"
    )



import datetime

import pytest

from px import px_loginhistory

from typing import Set

TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo


@pytest.fixture
def check_output(capfd):
    yield None

    out, err = capfd.readouterr()
    assert not err
    assert not out


def get_users_at(
    last_output: str, now: datetime.datetime, testtime: datetime.datetime
) -> Set[str]:
    """
    Ask px_loginhistory to parse last_output given the current timestamp of now.

    Then return the users px_loginhistory claims were logged in at testtime.
    """
    return px_loginhistory.get_users_at(testtime, last_output=last_output, now=now)


def test_get_users_at_range(check_output):
    # Test user logged in between two timestamps
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = "johan     ttys000                   Thu Mar 31 14:39 - 11:08  (20:29)"

    # Before
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 31, 14, 38, tzinfo=TIMEZONE),
    )

    # During
    assert {"johan"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 31, 14, 39, tzinfo=TIMEZONE),
    )
    assert {"johan"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 31, 17, 46, tzinfo=TIMEZONE),
    )
    assert {"johan"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 1, 11, 8, tzinfo=TIMEZONE),
    )

    # After
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 1, 11, 9, tzinfo=TIMEZONE),
    )


def test_get_users_at_range_raspbian(check_output):
    # Test user logged in between two timestamps, test string is from a Raspbian
    # system: https://github.com/walles/px/issues/130
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = (
        "sk-tv                                  Thu Mar 31 14:39 - down   (20:29)"
    )

    # Before
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 31, 14, 38, tzinfo=TIMEZONE),
    )

    # During
    assert {"sk-tv"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 31, 14, 39, tzinfo=TIMEZONE),
    )
    assert {"sk-tv"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 31, 17, 46, tzinfo=TIMEZONE),
    )
    assert {"sk-tv"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 1, 11, 8, tzinfo=TIMEZONE),
    )

    # After
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 1, 11, 9, tzinfo=TIMEZONE),
    )


def test_get_users_at_still_logged_in(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = "johan     ttys000                   Sun Apr  3 11:54   still logged in"

    # Before
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 3, 11, 53, tzinfo=TIMEZONE),
    )

    # During
    assert {"johan"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 3, 11, 54, tzinfo=TIMEZONE),
    )
    assert {"johan"} == get_users_at(lastline, now, datetime.datetime.now(TIMEZONE))


def test_get_users_at_remote(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = (
        "root     pts/1        10.1.6.120       Tue Jan 28 05:59   still logged in"
    )

    assert {"root from 10.1.6.120"} == get_users_at(
        lastline, now, datetime.datetime.now(TIMEZONE)
    )


def test_get_users_at_local_osx(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = "johan     ttys000                   Sun Apr  3 11:54   still logged in"

    assert {"johan"} == get_users_at(lastline, now, datetime.datetime.now(TIMEZONE))


def test_get_users_at_local_linux(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = (
        "johan    pts/2        :0               Wed Mar  9 13:25 - 13:38  (00:12)"
    )

    assert {"johan from :0"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 9, 13, 26, tzinfo=TIMEZONE),
    )


def test_get_users_at_until_crash(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = "johan     ttys001                   Thu Nov 26 19:55 - crash (27+07:11)"

    # Before
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2015, 11, 26, 19, 54, tzinfo=TIMEZONE),
    )

    # During
    assert {"johan"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2015, 11, 26, 19, 55, tzinfo=TIMEZONE),
    )
    assert {"johan"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2015, 12, 10, 19, 53, tzinfo=TIMEZONE),
    )

    # A bit after
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2015, 12, 26, 19, 55, tzinfo=TIMEZONE),
    )


def test_get_users_at_until_shutdown_osx(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = (
        "_mbsetupuser  console                   Mon Jan 18 20:31 - shutdown (34+01:29)"
    )

    # Before
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 1, 18, 20, 30, tzinfo=TIMEZONE),
    )

    # During
    assert {"_mbsetupuser"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 1, 18, 20, 31, tzinfo=TIMEZONE),
    )
    assert {"_mbsetupuser"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 2, 18, 20, 30, tzinfo=TIMEZONE),
    )

    # A bit after
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 2, 28, 20, 30, tzinfo=TIMEZONE),
    )


def test_get_users_at_until_shutdown_linux(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = (
        "johan    :0           :0               Sat Mar 26 22:04 - down   (00:08)"
    )

    # Before
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 26, 22, 3, tzinfo=TIMEZONE),
    )

    # During
    assert {"johan from :0"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 26, 22, 4, tzinfo=TIMEZONE),
    )
    assert {"johan from :0"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 26, 22, 9, tzinfo=TIMEZONE),
    )

    # A bit after
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 26, 22, 15, tzinfo=TIMEZONE),
    )


def test_get_users_at_multiple(check_output):
    # Test multiple users logged in between two timestamps
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    lastline = "\n".join(
        [
            "johan1     ttys000                   Thu Mar 31 14:39 - 11:08  (20:29)",
            "johan2     ttys000                   Thu Mar 31 14:39 - 11:08  (20:29)",
        ]
    )

    # Before
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 31, 14, 38, tzinfo=TIMEZONE),
    )

    # During
    assert {"johan1", "johan2"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 31, 14, 39, tzinfo=TIMEZONE),
    )
    assert {"johan1", "johan2"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 31, 17, 46, tzinfo=TIMEZONE),
    )
    assert {"johan1", "johan2"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 1, 11, 8, tzinfo=TIMEZONE),
    )

    # After
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 1, 11, 9, tzinfo=TIMEZONE),
    )


def test_get_users_at_pseudousers_osx(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)

    # Note trailing space in test string, we get that from last on OS X 10.11.3
    lastline = "reboot    ~                         Fri Oct 23 06:50 "
    # "reboot" is not a real user, it shouldn't be listed
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2015, 10, 23, 6, 50, tzinfo=TIMEZONE),
    )

    # Note trailing space in test string, we get that from last on OS X 10.11.3
    lastline = "shutdown  ~                         Fri Oct 23 06:49 "
    # "shutdown" is not a real user, it shouldn't be listed
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2015, 10, 23, 6, 49, tzinfo=TIMEZONE),
    )


def test_get_users_at_pseudousers_linux(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)

    lastline = (
        "reboot   system boot  4.2.0-30-generic Thu Mar  3 11:19 - 13:38 (6+02:18)"
    )
    # "reboot" is not a real user, it shouldn't be listed
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 3, 3, 11, 19, tzinfo=TIMEZONE),
    )


def test_get_users_at_gone_no_logout(check_output):
    """
    Treat "gone - no logout" as "still logged in".

    That's the only place I've seen it.
    """
    now = datetime.datetime(2016, 4, 7, 12, 8, tzinfo=TIMEZONE)
    lastline = (
        "johan    pts/3        :0               Mon Apr  4 23:10    gone - no logout"
    )

    # Before
    assert not get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 4, 23, 9, tzinfo=TIMEZONE),
    )

    # During
    assert {"johan from :0"} == get_users_at(
        lastline,
        now,
        datetime.datetime(2016, 4, 4, 23, 10, tzinfo=TIMEZONE),
    )
    assert {"johan from :0"} == get_users_at(
        lastline, now, datetime.datetime.now(TIMEZONE)
    )


def test_get_users_at_trailing_noise(check_output):
    now = datetime.datetime(2016, 4, 7, 12, 8, tzinfo=TIMEZONE)
    assert not get_users_at("", now, now)

    # Note trailing space in test string, we get that from last on OS X 10.11.3
    assert not get_users_at("wtmp begins Thu Oct  1 22:54 ", now, now)


def test_get_users_at_unexpected_last_output(caplog):
    UNEXPECTED = "glasskiosk"

    now = datetime.datetime(2016, 4, 7, 12, 8, tzinfo=TIMEZONE)
    assert not get_users_at(UNEXPECTED, now, now)

    assert UNEXPECTED in caplog.text


def test_get_users_at_just_run_it(check_output):
    # Just tyre kick it live wherever we happen to be. This shouldn't crash.
    px_loginhistory.get_users_at(datetime.datetime.now(TIMEZONE))


def test_to_timestamp(check_output):
    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    expected = datetime.datetime(2016, 3, 5, 11, 19, tzinfo=TIMEZONE)
    assert px_loginhistory._to_timestamp("Thu Mar  5 11:19", now) == expected

    now = datetime.datetime(2016, 4, 3, 12, 8, tzinfo=TIMEZONE)
    expected = datetime.datetime(2016, 2, 29, 13, 19, tzinfo=TIMEZONE)
    assert px_loginhistory._to_timestamp("Mon Feb 29 13:19", now) == expected

    now = datetime.datetime(2017, 1, 3, 12, 8, tzinfo=TIMEZONE)
    expected = datetime.datetime(2016, 2, 29, 13, 19, tzinfo=TIMEZONE)
    assert px_loginhistory._to_timestamp("Mon Feb 29 13:19", now) == expected


def test_to_timedelta(check_output):
    assert px_loginhistory._to_timedelta("01:29") == datetime.timedelta(
        0, hours=1, minutes=29
    )
    assert px_loginhistory._to_timedelta("4+01:29") == datetime.timedelta(
        4, hours=1, minutes=29
    )
    assert px_loginhistory._to_timedelta("34+01:29") == datetime.timedelta(
        34, hours=1, minutes=29
    )


def test_realworld_debian(check_output):
    """
    Regression test for https://github.com/walles/px/issues/48
    """
    now = datetime.datetime(2016, 12, 6, 9, 21, tzinfo=TIMEZONE)
    testtime = datetime.datetime(2016, 10, 24, 15, 34, tzinfo=TIMEZONE)
    lastline = (
        "norbert  pts/3        mosh [29846]     Wed Oct 24 15:33 - 15:34  (00:01)"
    )

    assert {"norbert from mosh"} == get_users_at(lastline, now, testtime)



#!/usr/bin/env python3

"""Benchmark loading the output of "lsof -F fnaptd0i" from a big system

Usage:
  benchmark_ipcmap.py <FILE>

FILE is a file containing the output of "lsof -F fnaptd0i".

This program will parse that output and make an IPC map of the process that has
the highest number of entries in that file.
"""

import os
import sys
import time

from typing import MutableMapping


MYDIR = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(MYDIR, ".."))

from tests import testutils  # noqa: E402
from px import px_file  # noqa: E402


# For how long should we do the benchmarking run (in seconds)
DURATION_S = 30


def get_most_common_pid(files):
    counts: MutableMapping[int, int] = {}
    for file in files:
        pid = file.pid
        if pid not in counts:
            counts[pid] = 0
        counts[pid] += 1
    return sorted(counts.keys(), key=lambda pid: counts[pid])[-1]


def get_timings(file, pid):
    """
    Loads file and creates an IPC map for PID.

    Returns timings in a tuple (load, mapping) in seconds.
    """
    t0 = time.time()
    files = None
    with open(file, encoding="utf-8") as lsof_output:
        files = px_file.lsof_to_files(lsof_output.read())
    t1 = time.time()
    dt_load = t1 - t0

    t0 = time.time()
    testutils.create_ipc_map(pid, files)
    t1 = time.time()
    dt_mapping = t1 - t0

    return (dt_load, dt_mapping)


def print_statistics(name, values):
    lowest = min(values)
    highest = max(values)
    middle = (lowest + highest) / 2
    radius = (highest - lowest) / 2
    print(f"{name} is {middle:.2f}s±{radius:.2f}s")


def main(lsof_file):
    print("Finding most popular PID...")
    files = None
    with open(lsof_file, encoding="utf-8") as lsof_output:
        files = px_file.lsof_to_files(lsof_output.read())
    pid = get_most_common_pid(files)
    print(f"Most popular PID: {pid}")

    end = time.time() + DURATION_S
    lap_number = 0
    load_times = []
    mapping_times = []
    total_times = []
    while time.time() < end:
        lap_number += 1
        print(f"Lap {lap_number}, {end - time.time():.0f}s left...")
        load_time, mapping_time = get_timings(lsof_file, pid)
        load_times.append(load_time)
        mapping_times.append(mapping_time)
        total_times.append(load_time + mapping_time)

    print_statistics("Loading time", load_times)
    print_statistics("Mapping time", mapping_times)
    print_statistics("  Total time", total_times)


if __name__ == "__main__":
    main(sys.argv[1])



from px import px
from px import px_process

from unittest.mock import patch

from typing import List

# FIXME: Distribute these tests to the other *_test.py files and remove this
# one.
#
# These tests all require Python 3. Back when we also supported Python 2 the
# tests in here used to live in their own Python 3 specific directory.
#
# But since we're now entirely on Python 3, I just moved the whole file in here.
# These tests should be moved into the correct files, and this file should be
# removed / renamed.


@patch("px.px.install")
def test_cmdline_install(mock):
    args = ["px", "--install"]
    px._main(args)
    mock.assert_called_once_with(args)


@patch("px.px_top.top")
def test_cmdline_top(mock):
    px._main(["px", "--top"])
    mock.assert_called_once()


@patch("px.px_top.top")
def test_cmdline_ptop(mock):
    px._main(["ptop"])
    mock.assert_called_once()


@patch("px.px_top.top")
def test_cmdline_top_with_search(mock):
    px._main(["px", "--top", "kalas"])

    mock.assert_called_once()
    args, kwargs = mock.call_args
    print(args)
    print(kwargs)

    assert kwargs.get("search") == "kalas"


def test_px_sort_cpupercent():
    px._main(["px", "--sort=cpupercent"])


@patch("px.px_top.top")
def test_cmdline_ptop_with_search(mock):
    px._main(["ptop", "kalas"])

    mock.assert_called_once()
    args, kwargs = mock.call_args
    print(args)
    print(kwargs)

    assert kwargs.get("search") == "kalas"


@patch("builtins.print")
def test_cmdline_help(mock):
    px._main(["px", "--help"])
    mock.assert_called_once_with(px.__doc__)


@patch("builtins.print")
def test_cmdline_version(mock):
    px._main(["px", "--version"])
    mock.assert_called_once()

    first_print_arg: str = mock.call_args.args[0]
    assert first_print_arg


@patch("px.px_processinfo.print_pid_info")
def test_cmdline_pid(mock):
    px._main(["px", "1235", "--no-pager"])

    mock.assert_called_once()

    print(mock)
    print(mock.call_args)
    args, kwargs = mock.call_args
    print(args)
    print(kwargs)

    # First arg is a file descriptor that changes on every invocation, check the
    # PID arg only.
    assert args[1] == 1235


@patch("px.px_terminal.to_screen_lines")
def test_cmdline_filter(mock):
    px._main(["px", "root", "--no-pager"])

    mock.assert_called_once()
    print(mock)
    print(mock.call_args)
    args, kwargs = mock.call_args
    print(args)
    print(kwargs)

    processes: List[px_process.PxProcess] = args[0]

    # This assumes something is running as root. We can assume that, right?
    assert len(processes) > 0

    # All listed processes must match
    for process in processes:
        assert process.match("root")


@patch("px.px_terminal.to_screen_lines")
def test_cmdline_list_all_processes(mock):
    px._main(["px"])

    mock.assert_called_once()
    args, kwargs = mock.call_args
    print(args)
    print(kwargs)
    processes: List[px_process.PxProcess] = args[0]

    # We are running, and something started us, so this list must not be empty
    assert len(processes) > 0
    assert processes[0].command



from . import testutils

from px import px_cwdfriends

from typing import List
from px import px_process


def test_current_cwd_unknown():
    process = testutils.create_process()
    test_me = px_cwdfriends.PxCwdFriends(process, [], [])
    assert test_me.cwd is None
    assert test_me.friends == []


def test_current_cwd_root():
    process = testutils.create_process(pid=123)
    cwd_file = testutils.create_file("xxx", "/", None, 123, fdtype="cwd")
    test_me = px_cwdfriends.PxCwdFriends(process, [process], [cwd_file])
    assert test_me.cwd == "/"
    assert test_me.friends == []


def test_current_cwd_notroot():
    process = testutils.create_process(pid=123)
    cwd_file = testutils.create_file("xxx", "/notroot", None, 123, fdtype="cwd")
    test_me = px_cwdfriends.PxCwdFriends(process, [process], [cwd_file])
    assert test_me.cwd == "/notroot"
    assert test_me.friends == []


def test_find_friends():
    process = testutils.create_process(pid=123)
    friend = testutils.create_process(pid=234)
    notfriend1 = testutils.create_process(pid=666)
    notfriend2 = testutils.create_process(pid=667)

    cwd_file = testutils.create_file("xxx", "/notroot", None, 123, fdtype="cwd")
    friend_cwd = testutils.create_file("xxx", "/notroot", None, 234, fdtype="cwd")
    notfriend_cwd = testutils.create_file(
        "xxx", "/somewhereelse", None, 666, fdtype="cwd"
    )
    # notfriend2 cwd intentionally left blank for the sake of this test

    test_me = px_cwdfriends.PxCwdFriends(
        process,
        [process, friend, notfriend1, notfriend2],
        [cwd_file, friend_cwd, notfriend_cwd],
    )

    assert test_me.cwd == "/notroot"
    assert test_me.friends == [friend]


def _get_friends_in_order(*args: str) -> List[str]:
    procs = []
    files = []
    for index, arg in enumerate(args):
        procs.append(testutils.create_process(pid=index, commandline=arg))
        files.append(testutils.create_file("xxx", "yyy", None, index, fdtype="cwd"))

    me = testutils.create_process(pid=1234, commandline="base-process")
    procs.append(me)
    files.append(testutils.create_file("xxx", "yyy", None, me.pid, fdtype="cwd"))

    return_me = []
    for friend in px_cwdfriends.PxCwdFriends(me, procs, files).friends:
        return_me.append(friend.cmdline)

    return return_me


def _get_friend_processes_in_order(
    *args: px_process.PxProcess,
) -> List[px_process.PxProcess]:
    files = []
    procs = list(args)
    for proc in procs:
        files.append(testutils.create_file("xxx", "yyy", None, proc.pid, fdtype="cwd"))

    me = testutils.create_process(pid=1234, commandline="base-process")
    procs.append(me)
    files.append(testutils.create_file("xxx", "yyy", None, me.pid, fdtype="cwd"))

    return px_cwdfriends.PxCwdFriends(me, procs, files).friends


def test_friend_ordering():
    # Test alphabetic order
    assert _get_friends_in_order("a", "b") == ["a", "b"]
    assert _get_friends_in_order("b", "a") == ["a", "b"]

    # Test ignoring initial -
    assert _get_friends_in_order("a", "-b", "c") == ["a", "-b", "c"]

    # Test PID as secondary key
    p1 = testutils.create_process(pid=1, commandline="a")
    p2 = testutils.create_process(pid=2, commandline="a")
    assert _get_friend_processes_in_order(p1, p2) == [p1, p2]
    assert _get_friend_processes_in_order(p2, p1) == [p1, p2]



from px import px_category_bar
from px import px_terminal


def test_render_bar_happy_path():
    names_and_numbers = [("apa", 1000.0), ("bepa", 300.0), ("cepa", 50.0)] + [
        ("long tail", 1)
    ] * 300
    assert px_category_bar.render_bar(10, names_and_numbers) == (
        px_terminal.red(" apa  ")
        + px_terminal.yellow(" b")
        + px_terminal.blue(" ")
        + px_terminal.inverse_video(" ")
    )


def test_render_bar_happy_path_unicode():
    names_and_numbers = [("åpa", 1000.0), ("bäpa", 300.0), ("cäpa", 50.0)] + [
        ("lång svans", 1)
    ] * 300
    assert px_category_bar.render_bar(10, names_and_numbers) == (
        px_terminal.red(" åpa  ")
        + px_terminal.yellow(" b")
        + px_terminal.blue(" ")
        + px_terminal.inverse_video(" ")
    )



#!/usr/bin/env python3

"""Benchmark listing all processes

Usage:
  benchmark_proc_get_all.py
"""

import os
import sys
import time


MYDIR = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(MYDIR, ".."))

from px import px_process  # noqa: E402

LAPS = 20


def main():
    t0 = time.time()
    for _ in range(LAPS):
        px_process.get_all()
    t1 = time.time()
    dt_seconds = t1 - t0

    print(f"Getting all processes takes {1000 * dt_seconds / LAPS:.0f}ms")


if __name__ == "__main__":
    main()



from px import px_units


def test_bytes_to_strings():
    assert px_units.bytes_to_strings(0, 1024 * 7 - 1) == ("0B", "7167B")
    assert px_units.bytes_to_strings(0, 1024 * 7) == ("0KB", "7KB")

    assert px_units.bytes_to_strings(0, (1024**2) * 7 - 1) == ("0KB", "7168KB")
    assert px_units.bytes_to_strings(0, (1024**2) * 7) == ("0MB", "7MB")

    assert px_units.bytes_to_strings(0, (1024**3) * 7 - 1) == ("0MB", "7168MB")
    assert px_units.bytes_to_strings(0, (1024**3) * 7) == ("0GB", "7GB")

    assert px_units.bytes_to_strings(0, (1024**4) * 7 - 1) == ("0GB", "7168GB")
    assert px_units.bytes_to_strings(0, (1024**4) * 7) == ("0TB", "7TB")

    assert px_units.bytes_to_strings(0, (1024**5) * 10) == ("0TB", "10240TB")



import os

from px import px_terminal

from . import testutils

from typing import List


def test_to_screen_lines_unbounded():
    px_terminal._enable_color = True
    procs = [testutils.create_process(commandline="/usr/bin/fluff 1234")]
    assert px_terminal.to_screen_lines(procs, None, None) == [
        "\x1b[1m  PID COMMAND USERNAME CPU CPUTIME RAM COMMANDLINE\x1b[22m",
        "47536 fluff   "
        + px_terminal.faint("root")
        + "     "
        + px_terminal.faint(" 0%")
        + " "
        + px_terminal.bold("  0.03s")
        + " "
        + px_terminal.faint(" 0%")
        + " /usr/bin/fluff 1234",
    ]


def test_to_screen_lines_unicode():
    px_terminal._enable_color = False
    procs = [testutils.create_process(commandline="/usr/bin/😀")]
    converted = px_terminal.to_screen_lines(procs, None, None)
    assert converted == [
        r"  PID COMMAND USERNAME CPU CPUTIME RAM COMMANDLINE",
        r"47536 😀       root      0%   0.03s  0% /usr/bin/😀",
    ]


def test_get_string_of_length():
    px_terminal._enable_color = True
    CSI = "\x1b["

    assert px_terminal.get_string_of_length("12345", 3) == "123"
    assert px_terminal.get_string_of_length("12345", 5) == "12345"
    assert px_terminal.get_string_of_length("12345", 7) == "12345  "
    assert px_terminal.get_string_of_length("12345", None) == "12345"

    # Test with escaped strings
    mid_bold = "1" + px_terminal.bold("234") + "5"
    assert px_terminal.get_string_of_length(mid_bold, 6) == mid_bold + " "
    assert (
        px_terminal.get_string_of_length(mid_bold, 3) == "1" + CSI + "1m23" + CSI + "0m"
    )


def test_crop_ansi_string_at_length():
    px_terminal._enable_color = True
    CSI = "\x1b["

    # Non-formatted cropping
    assert px_terminal.crop_ansi_string_at_length("", 0) == ""
    assert px_terminal.crop_ansi_string_at_length("", 4) == ""
    assert px_terminal.crop_ansi_string_at_length("1234", 4) == "1234"
    assert px_terminal.crop_ansi_string_at_length("1234", 3) == "123"
    assert px_terminal.crop_ansi_string_at_length("1234", 5) == "1234"

    bold_end = "1234" + px_terminal.bold("5678")
    assert px_terminal.crop_ansi_string_at_length(bold_end, 3) == "123"
    assert px_terminal.crop_ansi_string_at_length(bold_end, 4) == "1234"

    # Note that we expect the string cutter to add an end-marker when
    # breaking inside a formatted section
    assert px_terminal.crop_ansi_string_at_length(bold_end, 5) == "1234[1m5[0m".replace(
        "[", CSI
    )

    assert px_terminal.crop_ansi_string_at_length(
        bold_end, 8
    ) == "1234[1m5678[0m".replace("[", CSI)

    bold_middle = "123" + px_terminal.bold("456") + "789"
    assert px_terminal.crop_ansi_string_at_length(
        bold_middle, 6
    ) == "123[1m456[0m".replace("[", CSI)
    assert px_terminal.crop_ansi_string_at_length(
        bold_middle, 7
    ) == "123[1m456[22m7[0m".replace("[", CSI)


def test_getch():
    pipe = os.pipe()
    read, write = pipe
    os.write(write, b"q")

    # We should get unicode responses from getch()
    sequence = px_terminal.getch(timeout_seconds=0, fd=read)
    assert sequence is not None
    assert sequence._string == "q"


def test_tokenize():
    px_terminal._enable_color = True
    tokenize_me = "ab" + px_terminal.bold("c") + "de"
    parts: List[str] = []
    for token in px_terminal._tokenize(tokenize_me):
        parts.append(token)

    assert parts == ["ab", "\x1b[1m", "c", "\x1b[22m", "de"]



import os
import sys

from px import px


def test_run_on_pid(capfd):
    """
    Just run px on a PID.

    The only verification done here is that it doesn't crash,
    there is room for improvement...
    """
    argv = [
        sys.argv[0],
        "--no-pager",  # Paging causes problems on Travis CI
        # Note that px hides our own PID by design, so we look for our
        # parent PID in this test.
        str(os.getppid()),
    ]

    # Enable manual inspection of the output:
    # https://docs.pytest.org/en/latest/capture.html
    with capfd.disabled():
        px._main(argv)



import getpass
import datetime

import os
import pytest

from px import px_process
from . import testutils

from typing import MutableSet


def test_create_process():
    process_builder = px_process.PxProcessBuilder()
    process_builder.pid = 7
    process_builder.ppid = 1
    process_builder.rss_kb = 123
    process_builder.start_time_string = testutils.TIMESTRING
    process_builder.username = "usernamex"
    process_builder.cpu_time = 1.3
    process_builder.memory_percent = 42.7
    process_builder.cmdline = "hej kontinent"
    test_me = process_builder.build(testutils.local_now())

    assert test_me.pid == 7
    assert test_me.ppid == 1
    assert test_me.rss_kb == 123
    assert test_me.username == "usernamex"
    assert test_me.cpu_time_s == "1.3s"
    assert test_me.memory_percent_s == "43%"
    assert test_me.cmdline == "hej kontinent"

    assert test_me.start_time == testutils.TIME
    assert test_me.age_seconds > 0


def test_create_future_process():
    """
    Handle the case where we first look at the clock, then list processes.

    Let's say that:
    1. We look at the clock, and the clock is 12:34:56
    2. We find a newly started process at     12:34:57

    This case used to lead to crashes when we asserted for this:
    https://github.com/walles/px/issues/84
    """
    process_builder = px_process.PxProcessBuilder()

    # This is what we want to test
    process_builder.start_time_string = testutils.TIMESTRING
    before_the_process_was_started = testutils.TIME - datetime.timedelta(seconds=1)

    # These values are required to not fail in other ways
    process_builder.cmdline = "hej kontinent"
    process_builder.pid = 1
    process_builder.rss_kb = 123
    process_builder.username = "johan"

    # Test it!
    test_me = process_builder.build(before_the_process_was_started)
    assert test_me.age_seconds == 0


def test_ps_line_to_process_unicode():
    process = testutils.create_process(cputime="2:14.15")

    assert process.username == "root"
    assert process.cmdline == "/usr/sbin/cupsd -l"


def test_ps_line_to_process_1():
    process = testutils.create_process()

    assert process.pid == 47536
    assert process.ppid == 1234
    assert process.username == "root"
    assert process.cpu_time_s == "0.03s"
    assert process.memory_percent_s == "0%"
    assert process.cmdline == "/usr/sbin/cupsd -l"

    assert process.start_time == testutils.TIME
    assert process.age_seconds > 0


def test_ps_line_to_process_2():
    process = testutils.create_process(cputime="2:14.15")

    assert process.pid == 47536
    assert process.ppid == 1234
    assert process.username == "root"
    assert process.cpu_time_s == "2m14s"
    assert process.memory_percent_s == "0%"
    assert process.cmdline == "/usr/sbin/cupsd -l"

    assert process.start_time == testutils.TIME
    assert process.age_seconds > 0


# From a real-world failure
def test_ps_line_to_process_3():
    process = px_process.ps_line_to_process(
        "  5328"
        "   4432"
        "   123"
        " Thu Feb 25 07:42:36 2016"
        " " + str(os.getuid()) + " 5.5"
        "    1-19:31:31"
        " 19.7"
        " /usr/sbin/mysqld"
        " --basedir=/usr"
        " --datadir=/data/user/mysql"
        " --plugin-dir=/usr/lib/mysql/plugin"
        " --user=mysql"
        " --log-error=/var/log/mysql/mysql.err"
        " --pid-file=/var/run/mysqld/mysqld.pid"
        " --socket=/var/run/mysqld/mysqld.sock"
        " --port=3306",
        testutils.local_now(),
    )
    assert process.username == getpass.getuser()
    assert process.cpu_percent_s == "6%"
    assert process.memory_percent_s == "20%"
    assert process.cpu_time_s == "1d19h"
    assert process.command == "mysqld"


def _validate_references(processes):
    """Fsck the parent / children relationships between all processes"""
    for process in processes:
        if process.pid == 0:
            assert process.parent is None
        else:
            assert process.parent is not None

        assert isinstance(process.children, list)
        if process.parent:
            assert process.parent in processes
            assert process.parent.pid == process.ppid
            assert process in process.parent.children

        for child in process.children:
            assert child in processes
            assert child.parent == process


def _test_get_all():
    all_processes = px_process.get_all()
    assert len(all_processes) >= 4  # Expect at least kernel, init, bash and python
    for process in all_processes:
        assert process is not None

    pids = list(map(lambda p: p.pid, all_processes))

    # Finding ourselves is just confusing...
    assert os.getpid() not in pids

    # ... but all other processes should be there
    assert os.getppid() in pids

    # PID 1 is launchd on OS X, init on Linux.
    #
    # If there's a system where PID 1 doesn't exist this test needs to be modded
    # and that system documented here.
    assert 1 in pids

    # Assert that all contains no duplicate PIDs
    seen_pids: MutableSet[int] = set()
    for process in all_processes:
        pid = process.pid
        assert pid not in seen_pids
        seen_pids.add(pid)

    # Assert that there are processes with the current user name
    current_users_processes = filter(
        lambda process: process.username == getpass.getuser(), all_processes
    )
    assert current_users_processes

    _validate_references(all_processes)

    for process in all_processes:
        assert isinstance(process.cmdline, str)
        assert isinstance(process.username, str)

    if os.environ.get("CI") == "true":
        # Checking for future processes sometimes fails in CI. Since I don't
        # understand it and can't fix it, let's just not look for that in CI. If
        # it happens locally, then that's a good start for troupleshooting.
        #
        # For reference, in
        # https://github.com/Homebrew/homebrew-core/pull/186101 four different
        # runs of "macOS 14-arm64" failed like this, all with processes created
        # around 100s-210s in the future.
        return

    # Ensure no processes are from the future
    now = testutils.local_now()
    for process in all_processes:
        # Processes created in the future = fishy
        assert process.age_seconds >= 0
        assert process.start_time < now


def test_get_all_swedish():
    """
    In Swedish, floating point numbers are indicated with comma, so 4.2 in
    English is 4,2 in Swedish. This test verifies that setting a Swedish locale
    won't mess up our parsing.
    """
    os.environ["LANG"] = "sv_SE.UTF-8"
    os.environ["LC_TIME"] = "sv_SE.UTF-8"
    os.environ["LC_NUMERIC"] = "sv_SE.UTF-8"

    _test_get_all()


def test_get_all_defaultlocale():
    del os.environ["LANG"]
    _test_get_all()


def test_process_eq():
    """Compare two mostly identical processes, where one has a parent and the other one not"""
    process_a = testutils.create_process()

    process_b = testutils.create_process()
    parent = px_process.create_kernel_process(testutils.local_now())
    process_b.parent = parent

    assert process_a != process_b


def test_parse_time():
    assert px_process.parse_time("0:00.03") == 0.03
    assert px_process.parse_time("1:02.03") == 62.03
    assert px_process.parse_time("03:35:32") == 3 * 60 * 60 + 35 * 60 + 32
    assert px_process.parse_time("9-03:35:32") == 9 * 86400 + 3 * 60 * 60 + 35 * 60 + 32

    with pytest.raises(ValueError) as e:
        px_process.parse_time("Constantinople")
    assert "Constantinople" in str(e.value)


def test_match():
    p = testutils.create_process(uid=0, commandline="/usr/libexec/AirPlayXPCHelper")

    assert p.match(None)

    assert p.match("root")
    assert not p.match("roo")

    assert p.match("Air")
    assert p.match("Play")

    assert p.match("air")
    assert p.match("play")

    # Match PID by prefix but not substring. Exact matches are used for
    # searching in ptop. Prefix matching is used to not throw the right answer
    # away while the user is typing their search in ptop. Substring matching has
    # no value.
    assert p.match("47536")
    assert p.match("4753")
    assert not p.match("7536")


def test_seconds_to_str():
    assert px_process.seconds_to_str(0.54321) == "0.54s"
    assert px_process.seconds_to_str(1.54321) == "1.54s"
    assert px_process.seconds_to_str(0.5) == "0.5s"
    assert px_process.seconds_to_str(1.0) == "1.0s"
    assert px_process.seconds_to_str(1) == "1s"

    assert px_process.seconds_to_str(60.54321) == "1m00s"

    assert px_process.seconds_to_str(3598.54321) == "59m58s"
    assert px_process.seconds_to_str(3659.54321) == "1h00m"
    assert px_process.seconds_to_str(3660.54321) == "1h01m"
    assert px_process.seconds_to_str(4260.54321) == "1h11m"

    t1h = 3600
    t1d = t1h * 24
    assert px_process.seconds_to_str(t1d + 3598) == "1d00h"
    assert px_process.seconds_to_str(t1d + 3659) == "1d01h"
    assert px_process.seconds_to_str(t1d + t1h * 11) == "1d11h"


def test_get_command_line_array():
    p = testutils.create_process(commandline="/usr/libexec/AirPlayXPCHelper")
    assert p.get_command_line_array() == ["/usr/libexec/AirPlayXPCHelper"]

    p = testutils.create_process(commandline="/usr/sbin/universalaccessd launchd -s")
    assert p.get_command_line_array() == ["/usr/sbin/universalaccessd", "launchd", "-s"]


def test_get_command_line_array_space_in_binary(tmpdir):
    # Create a file name with a space in it
    spaced_path = tmpdir.join("i contain spaces")
    spaced_path.write_binary(b"")
    spaced_name = str(spaced_path)

    # Verify splitting of the spaced file name
    p = testutils.create_process(commandline=spaced_name)
    assert p.get_command_line_array() == [spaced_name]

    # Verify splitting with more parameters on the line
    p = testutils.create_process(commandline=spaced_name + " parameter")
    assert p.get_command_line_array() == [spaced_name, "parameter"]


def test_command_dotted_prefix():
    # If there's a dot with a lot of text after it we should drop everything
    # before the dot.
    p = testutils.create_process(
        commandline="/.../com.apple.InputMethodKit.TextReplacementService"
    )
    assert p.command == "TextReplacementService"

    # If there's a dot with four characters or less after it, assume it's a file
    # suffix and take the next to last section
    p = testutils.create_process(
        commandline="/.../com.apple.InputMethodKit.TextReplacementService.1234"
    )
    assert p.command == "TextReplacementService"
    p = testutils.create_process(
        commandline="/.../com.apple.InputMethodKit.TextReplacementService.12345"
    )
    assert p.command == "12345"


def test_command_linux_kernelproc():
    p = testutils.create_process(commandline="[ksoftirqd/0]")
    assert p.command == "[ksoftirqd/0]"

    p = testutils.create_process(commandline="[kworker/0:0H]")
    assert p.command == "[kworker/0:0H]"

    p = testutils.create_process(commandline="[rcuob/3]")
    assert p.command == "[rcuob/3]"


def test_command_in_parentheses():
    # Observed on OS X
    p = testutils.create_process(commandline="(python2.7)")
    assert p.command == "(python2.7)"


def test_uid_to_username():
    username = px_process.uid_to_username(os.getuid())
    assert username == getpass.getuser()
    assert isinstance(username, str)

    username = px_process.uid_to_username(456789)
    assert username == "456789"
    assert isinstance(username, str)


def test_resolve_links():
    UNKNOWN_PID = 1323532
    p1 = testutils.create_process(pid=1, ppid=UNKNOWN_PID)
    p2 = testutils.create_process(pid=2, ppid=1)
    processes = {p1.pid: p1, p2.pid: p2}
    px_process.resolve_links(processes, testutils.local_now())

    assert p1.parent is None
    assert p2.parent is p1

    # Verify both equality...
    assert p1.children == [p2]
    # ... and identity of the child.
    assert list(p1.children)[0] is p2


def test_resolve_links_multiple_roots():
    """There can be only one. Root. Of the process tree."""
    processes = px_process.get_all()
    process_map = {p.pid: p for p in processes}
    px_process.resolve_links(process_map, testutils.local_now())

    root0 = processes[0]
    while root0.parent:
        root0 = root0.parent

    for process in processes:
        root = process
        while root.parent:
            root = root.parent

        assert root is root0



#!/usr/bin/env python3

import os
import time
import signal


def receiveSignal(signalNumber, _):
    print("Received signal " + str(signalNumber))


signal.signal(signal.SIGTERM, receiveSignal)

print('Waiting to be killed using "kill -9 ' + str(os.getpid()) + '"...')
while True:
    time.sleep(1234)



from px import px_process
from px import px_ipc_map
from px import px_terminal
from px import px_processinfo

from . import testutils

import sys

from typing import List, Tuple


def test_to_relative_start_string():
    base = testutils.create_process(pid=100, timestring="Mon Mar  7 09:33:11 2016")
    close = testutils.create_process(pid=101, timestring="Mon Mar  7 09:33:12 2016")
    assert (
        "cupsd(101) was started 1.0s after cupsd(100)"
        == px_processinfo.to_relative_start_string(base, close)
    )

    base = testutils.create_process(pid=100, timestring="Mon Mar  7 09:33:11 2016")
    close = testutils.create_process(pid=101, timestring="Mon Mar  7 09:33:10 2016")
    assert (
        "cupsd(101) was started 1.0s before cupsd(100)"
        == px_processinfo.to_relative_start_string(base, close)
    )

    base = testutils.create_process(pid=100, timestring="Mon Mar  7 09:33:11 2016")
    close = testutils.create_process(pid=101, timestring="Mon Mar  7 09:33:11 2016")
    assert (
        "cupsd(101) was started just after cupsd(100)"
        == px_processinfo.to_relative_start_string(base, close)
    )


def test_get_closests_starts_all_within_1s():
    # Verify that even if we have a large number of processes created within 1s
    # of the base one, we get all of those
    all_procs = []
    all_procs.append(
        testutils.create_process(pid=100, timestring="Mon Mar  7 09:33:09 2016")
    )
    all_procs.append(
        testutils.create_process(pid=101, timestring="Mon Mar  7 09:33:10 2016")
    )
    all_procs.append(
        testutils.create_process(pid=102, timestring="Mon Mar  7 09:33:10 2016")
    )
    all_procs.append(
        testutils.create_process(pid=103, timestring="Mon Mar  7 09:33:10 2016")
    )
    all_procs.append(
        testutils.create_process(pid=104, timestring="Mon Mar  7 09:33:11 2016")
    )

    base = testutils.create_process(pid=105, timestring="Mon Mar  7 09:33:11 2016")
    all_procs.append(base)

    all_procs.append(
        testutils.create_process(pid=106, timestring="Mon Mar  7 09:33:11 2016")
    )
    all_procs.append(
        testutils.create_process(pid=107, timestring="Mon Mar  7 09:33:12 2016")
    )
    all_procs.append(
        testutils.create_process(pid=108, timestring="Mon Mar  7 09:33:12 2016")
    )
    all_procs.append(
        testutils.create_process(pid=109, timestring="Mon Mar  7 09:33:12 2016")
    )
    all_procs.append(
        testutils.create_process(pid=110, timestring="Tue Mar  8 09:33:13 2016")
    )

    close = px_processinfo.get_closest_starts(base, all_procs)
    assert len(close) == 8
    assert all_procs[0] not in close
    assert all_procs[1] in close
    assert all_procs[2] in close
    assert all_procs[3] in close
    assert all_procs[4] in close
    assert all_procs[5] not in close  # This is base, it shouldn't be close to itself
    assert all_procs[6] in close
    assert all_procs[7] in close
    assert all_procs[8] in close
    assert all_procs[9] in close
    assert all_procs[10] not in close


def test_get_closest_starts_five_closest():
    # Verify that we list the five closest processes even if none of them are
    # very close
    all_procs = []
    all_procs.append(
        testutils.create_process(pid=102, timestring="Mon Mar  7 06:33:10 2016")
    )
    all_procs.append(
        testutils.create_process(pid=103, timestring="Mon Mar  7 07:33:10 2016")
    )
    all_procs.append(
        testutils.create_process(pid=104, timestring="Mon Mar  7 08:33:11 2016")
    )

    base = testutils.create_process(pid=105, timestring="Mon Mar  7 09:33:11 2016")
    all_procs.append(base)

    all_procs.append(
        testutils.create_process(pid=106, timestring="Mon Mar  7 10:33:11 2016")
    )
    all_procs.append(
        testutils.create_process(pid=107, timestring="Mon Mar  7 11:33:12 2016")
    )
    all_procs.append(
        testutils.create_process(pid=108, timestring="Mon Mar  7 11:43:12 2016")
    )
    all_procs.append(
        testutils.create_process(pid=110, timestring="Tue Mar  7 12:33:13 2016")
    )

    close = px_processinfo.get_closest_starts(base, all_procs)
    assert len(close) == 5
    assert all_procs[0] not in close
    assert all_procs[1] in close
    assert all_procs[2] in close
    assert all_procs[3] not in close  # This is base, it shouldn't be close to itself
    assert all_procs[4] in close
    assert all_procs[5] in close
    assert all_procs[6] in close
    assert all_procs[7] not in close


def test_print_starttime():
    # Just make sure it doesn't crash
    all_procs = px_process.get_all()
    process0 = list(filter(lambda p: p.pid == 0, all_procs))[0]
    px_processinfo.print_start_time(sys.stdout.fileno(), process0)


def test_print_process_subtree():
    lines: List[Tuple[str, px_process.PxProcess]] = []

    child_proc = testutils.create_process(pid=2, commandline="child")
    child_proc.children = []

    parent_proc = testutils.create_process(pid=1, commandline="parent")
    parent_proc.children = [child_proc]

    px_processinfo.print_process_subtree(sys.stdout.fileno(), parent_proc, 0, lines)

    assert lines == [
        ("" + str(parent_proc), parent_proc),
        ("  " + str(child_proc), child_proc),
    ]


def test_to_ipc_lines():
    ipcmap = {
        px_ipc_map.PeerProcess(name="foo", pid=47536): [
            testutils.create_file("PIPE", "[] ->0xAda", "0xE0e", 25),
        ],
        px_ipc_map.PeerProcess(name="bar", pid=47536): [
            testutils.create_file("PIPE", "[] ->0xAda", "0xE0e", 25),
        ],
    }
    lines = px_processinfo.to_ipc_lines(ipcmap)  # type: ignore
    assert lines == [
        px_terminal.bold("bar(47536)") + ": [PIPE] ->0xAda",
        px_terminal.bold("foo(47536)") + ": [PIPE] ->0xAda",
    ]
